from dataclasses import MISSING
from typing import Sequence

import torch

from srb import assets
from srb._typing import StepReturn
from srb.core.asset import (
    Articulation,
    AssetVariant,
    Object,
    RigidObject,
    RigidObjectCfg,
    RigidObjectCollection,
    RigidObjectCollectionCfg,
)
from srb.core.env import (
    ManipulationEnv,
    ManipulationEnvCfg,
    ManipulationEventCfg,
    ManipulationSceneCfg,
)
from srb.core.manager import EventTermCfg, SceneEntityCfg
from srb.core.marker import VisualizationMarkers, VisualizationMarkersCfg
from srb.core.mdp import (
    reset_collection_root_state_uniform_poisson_disk_2d,
    reset_root_state_uniform,
)
from srb.core.sensor import ContactSensor, ContactSensorCfg
from srb.core.sim import PreviewSurfaceCfg
from srb.tasks.manipulation.peg_in_hole_assembly.asset import (
    select_peg_in_hole_assembly,
)
from srb.utils.cfg import configclass
from srb.utils.math import (
    combine_frame_transforms,
    matrix_from_quat,
    rotmat_to_rot6d,
    scale_transform,
    subtract_frame_transforms,
)

from .asset import select_solar_panel

##############
### Config ###
##############


@configclass
class SceneCfg(ManipulationSceneCfg):
    peg: RigidObjectCollectionCfg = RigidObjectCollectionCfg(
        rigid_objects=MISSING,  # type: ignore
    )
    hole: RigidObjectCollectionCfg = RigidObjectCollectionCfg(
        rigid_objects=MISSING,  # type: ignore
    )
    panel: RigidObjectCfg = MISSING  # type: ignore


@configclass
class EventCfg(ManipulationEventCfg):
    randomize_object_state: EventTermCfg = EventTermCfg(
        func=reset_collection_root_state_uniform_poisson_disk_2d,
        mode="reset",
        params={
            "asset_cfg": SceneEntityCfg("peg"),
            "pose_range": {
                "x": (
                    -0.2,
                    0.05,
                ),
                "y": (
                    -0.05,
                    0.05,
                ),
                "z": (
                    0.025,
                    0.025,
                ),
                "roll": (torch.pi / 2, torch.pi / 2),
                "pitch": (-torch.pi, torch.pi),
                "yaw": (-torch.pi, torch.pi),
            },
            "velocity_range": {},
            "radius": 0.2,
        },
    )
    randomize_panel_state: EventTermCfg = EventTermCfg(
        func=reset_root_state_uniform,
        mode="reset",
        params={
            "asset_cfg": SceneEntityCfg("panel"),
            "pose_range": {
                "x": (-0.1 - 0.025, -0.1 + 0.025),
                "y": (-0.45 - 0.025, -0.45 + 0.025),
                "z": (-0.14, -0.14),
                "roll": (torch.pi, torch.pi),
                "pitch": (-torch.pi + torch.pi / 7, -torch.pi + torch.pi / 7),
                "yaw": (
                    torch.pi / 2 - torch.pi / 32,
                    torch.pi / 2 + torch.pi / 32,
                ),
            },
            "velocity_range": {},
        },
    )


@configclass
class TaskCfg(ManipulationEnvCfg):
    ## Assets
    peg: Object | AssetVariant = AssetVariant.DATASET

    ## Scene
    scene: SceneCfg = SceneCfg()

    ## Events
    events: EventCfg = EventCfg()

    ## Time
    episode_length_s: float = 50.0
    is_finite_horizon: bool = True

    ## Target - Panel
    panel_tf_pos_target = (0.55, 0.0, 0.0)
    panel_tf_quat_target = (1.0, 0.0, 0.0, 0.0)
    panel_target_marker_cfg = VisualizationMarkersCfg(
        prim_path="/Visuals/panel_target",
        markers={
            "target": assets.SolarPanel().asset_cfg.spawn.replace(  # type: ignore
                visual_material=PreviewSurfaceCfg(emissive_color=(0.2, 0.2, 0.8)),
            )
        },
    )

    def __post_init__(self):
        super().__post_init__()

        # Task setup: Peg-in-hole assembly
        self.peg_in_hole_assemblies = [
            select_peg_in_hole_assembly(
                self,  # type: ignore
                seed=self.seed,
                prim_path_peg=f"{{ENV_REGEX_NS}}/peg{i}",
                prim_path_hole=f"{{ENV_REGEX_NS}}/hole{i}",
                init_state=RigidObjectCfg.InitialStateCfg(pos=init_pos),
                short_peg=short_peg,
                peg_kwargs={
                    "activate_contact_sensors": True,
                },
            )
            for i, (init_pos, short_peg) in enumerate(
                [
                    ((0.55 + 0.1, 0.2, 0.04), True),
                    ((0.55 + 0.1, -0.2, 0.04), True),
                    ((0.55 - 0.1, 0.2, 0.04), False),
                    ((0.55 - 0.1, -0.2, 0.04), False),
                ]
            )
        ]
        # Scene: Peg
        self.scene.peg.rigid_objects = {
            f"peg{i}": cfg.peg.asset_cfg.replace(  # type: ignore
                init_state=RigidObjectCfg.InitialStateCfg(
                    pos=(0.5, 0.0, 0.13),
                )
            )
            for i, cfg in enumerate(self.peg_in_hole_assemblies)
        }
        # Scene: Hole
        self.scene.hole.rigid_objects = {
            f"hole{i}": cfg.hole.asset_cfg
            for i, cfg in enumerate(self.peg_in_hole_assemblies)
        }
        # Scene: Panel
        self.panel_cfg = select_solar_panel(
            self,
            init_state=RigidObjectCfg.InitialStateCfg(pos=(0.55, 0.0, 0.04)),
        )
        self.scene.panel = self.panel_cfg.asset_cfg
        self.panel_tf_pos_target = (
            self.panel_tf_pos_target[0] + self.panel_cfg.offset_pos[0],
            self.panel_tf_pos_target[1] + self.panel_cfg.offset_pos[1],
            self.panel_tf_pos_target[2] + self.panel_cfg.offset_pos[2] + 0.04,
        )

        # Sensor: End-effector contacts
        if isinstance(self.scene.contacts_end_effector, ContactSensorCfg):
            self.scene.contacts_end_effector.filter_prim_paths_expr = [
                f"{{ENV_REGEX_NS}}/peg{i}" for i in range(4)
            ]

        # Update seed & number of variants for procedural assets
        self._update_procedural_assets()


############
### Task ###
############


class Task(ManipulationEnv):
    cfg: TaskCfg

    def __init__(self, cfg: TaskCfg, **kwargs):
        super().__init__(cfg, **kwargs)

        ## Get scene assets
        self._objs: RigidObjectCollection = self.scene["peg"]
        self._targets: RigidObjectCollection = self.scene["hole"]
        self._panel: RigidObject = self.scene["panel"]
        self._panel_target_marker: VisualizationMarkers = VisualizationMarkers(
            self.cfg.panel_target_marker_cfg
        )

        ## Initialize buffers
        self._tf_pos_objs_initial = torch.zeros(
            (self.num_envs, 4, 3),
            dtype=torch.float32,
            device=self.device,
        )
        self._offset_pos_pegs_ends = torch.tensor(
            [self.cfg.peg_in_hole_assemblies[i].peg.offset_pos_ends for i in range(4)],
            dtype=torch.float32,
            device=self.device,
        ).repeat(self.num_envs, 1, 1, 1)

        self._symmetry_pegs = torch.tensor(
            [self.cfg.peg_in_hole_assemblies[i].peg.symmetry for i in range(4)],
            dtype=torch.int32,
            device=self.device,
        ).repeat(self.num_envs, 1)
        self._offset_pos_holes_bottom = torch.tensor(
            [
                self.cfg.peg_in_hole_assemblies[i].hole.offset_pos_bottom
                for i in range(4)
            ],
            dtype=torch.float32,
            device=self.device,
        ).repeat(self.num_envs, 1, 1)
        self._offset_pos_holes_entrance = torch.tensor(
            [
                self.cfg.peg_in_hole_assemblies[i].hole.offset_pos_entrance
                for i in range(4)
            ],
            dtype=torch.float32,
            device=self.device,
        ).repeat(self.num_envs, 1, 1)
        self._tf_pos_panel_initial = torch.zeros(
            (self.num_envs, 3), dtype=torch.float32, device=self.device
        )
        self._tf_pos_panel_target = self.scene.env_origins + torch.tensor(
            self.cfg.panel_tf_pos_target, dtype=torch.float32, device=self.device
        ).repeat(self.num_envs, 1)
        self._tf_quat_panel_target = torch.tensor(
            self.cfg.panel_tf_quat_target, dtype=torch.float32, device=self.device
        ).repeat(self.num_envs, 1)

        ## Visualize target
        self._panel_target_marker.visualize(
            self._tf_pos_panel_target, self._tf_quat_panel_target
        )

    def _reset_idx(self, env_ids: Sequence[int]):
        super()._reset_idx(env_ids)
        self._tf_pos_objs_initial[env_ids] = self._objs.data.object_pos_w[env_ids]
        self._tf_pos_panel_initial[env_ids] = self._panel.data.root_pos_w[env_ids]

    def extract_step_return(self) -> StepReturn:
        return _compute_step_return(
            ## Time
            episode_length=self.episode_length_buf,
            max_episode_length=self.max_episode_length,
            truncate_episodes=self.cfg.truncate_episodes,
            ## Actions
            act_current=self.action_manager.action,
            act_previous=self.action_manager.prev_action,
            ## States
            # Joints
            joint_pos_robot=self._robot.data.joint_pos,
            joint_pos_limits_robot=(
                self._robot.data.soft_joint_pos_limits
                if torch.all(torch.isfinite(self._robot.data.soft_joint_pos_limits))
                else None
            ),
            joint_pos_end_effector=self._end_effector.data.joint_pos
            if isinstance(self._end_effector, Articulation)
            else None,
            joint_pos_limits_end_effector=(
                self._end_effector.data.soft_joint_pos_limits
                if isinstance(self._end_effector, Articulation)
                and torch.all(
                    torch.isfinite(self._end_effector.data.soft_joint_pos_limits)
                )
                else None
            ),
            joint_acc_robot=self._robot.data.joint_acc,
            joint_applied_torque_robot=self._robot.data.applied_torque,
            # Kinematics
            fk_pos_end_effector=self._tf_end_effector.data.target_pos_source[:, 0, :],
            fk_quat_end_effector=self._tf_end_effector.data.target_quat_source[:, 0, :],
            # Transforms (world frame)
            tf_pos_end_effector=self._tf_end_effector.data.target_pos_w[:, 0, :],
            tf_quat_end_effector=self._tf_end_effector.data.target_quat_w[:, 0, :],
            tf_pos_objs_initial=self._tf_pos_objs_initial,
            tf_pos_objs=self._objs.data.object_pos_w,
            tf_quat_objs=self._objs.data.object_quat_w,
            tf_pos_targets=self._targets.data.object_pos_w,
            tf_quat_targets=self._targets.data.object_quat_w,
            tf_pos_panel_initial=self._tf_pos_panel_initial,
            tf_pos_panel=self._panel.data.root_com_pos_w,
            tf_quat_panel=self._panel.data.root_com_quat_w,
            tf_pos_panel_target=self._tf_pos_panel_target,
            tf_quat_panel_target=self._tf_quat_panel_target,
            # Contacts
            contact_forces_robot=self._contacts_robot.data.net_forces_w,  # type: ignore
            contact_forces_end_effector=self._contacts_end_effector.data.net_forces_w
            if isinstance(self._contacts_end_effector, ContactSensor)
            else None,
            contact_force_matrix_end_effector=self._contacts_end_effector.data.force_matrix_w
            if isinstance(self._contacts_end_effector, ContactSensor)
            else None,
            ## Peg-in-hole descriptors
            offset_pos_pegs_ends=self._offset_pos_pegs_ends,
            symmetry_peg=self._symmetry_pegs,
            offset_pos_holes_bottom=self._offset_pos_holes_bottom,
            offset_pos_holes_entrance=self._offset_pos_holes_entrance,
        )


@torch.jit.script
def _compute_step_return(
    *,
    ## Time
    episode_length: torch.Tensor,
    max_episode_length: int,
    truncate_episodes: bool,
    ## Actions
    act_current: torch.Tensor,
    act_previous: torch.Tensor,
    ## States
    # Joints
    joint_pos_robot: torch.Tensor,
    joint_pos_limits_robot: torch.Tensor | None,
    joint_pos_end_effector: torch.Tensor | None,
    joint_pos_limits_end_effector: torch.Tensor | None,
    joint_acc_robot: torch.Tensor,
    joint_applied_torque_robot: torch.Tensor,
    # Kinematics
    fk_pos_end_effector: torch.Tensor,
    fk_quat_end_effector: torch.Tensor,
    # Transforms (world frame)
    tf_pos_end_effector: torch.Tensor,
    tf_quat_end_effector: torch.Tensor,
    tf_pos_objs_initial: torch.Tensor,
    tf_pos_objs: torch.Tensor,
    tf_quat_objs: torch.Tensor,
    tf_pos_targets: torch.Tensor,
    tf_quat_targets: torch.Tensor,
    tf_pos_panel_initial: torch.Tensor,
    tf_pos_panel: torch.Tensor,
    tf_quat_panel: torch.Tensor,
    tf_pos_panel_target: torch.Tensor,
    tf_quat_panel_target: torch.Tensor,
    # Contacts
    contact_forces_robot: torch.Tensor,
    contact_forces_end_effector: torch.Tensor | None,
    contact_force_matrix_end_effector: torch.Tensor | None,
    ## Peg-in-hole descriptors
    offset_pos_pegs_ends: torch.Tensor,
    symmetry_peg: torch.Tensor,
    offset_pos_holes_bottom: torch.Tensor,
    offset_pos_holes_entrance: torch.Tensor,
) -> StepReturn:
    num_envs = episode_length.size(0)
    num_objs = tf_pos_objs.size(1)
    dtype = episode_length.dtype
    device = episode_length.device

    ############
    ## States ##
    ############
    ## Joints
    # Robot joints
    joint_pos_robot_normalized = (
        scale_transform(
            joint_pos_robot,
            joint_pos_limits_robot[:, :, 0],
            joint_pos_limits_robot[:, :, 1],
        )
        if joint_pos_limits_robot is not None
        else joint_pos_robot
    )
    # End-effector joints
    joint_pos_end_effector_normalized = (
        scale_transform(
            joint_pos_end_effector,
            joint_pos_limits_end_effector[:, :, 0],
            joint_pos_limits_end_effector[:, :, 1],
        )
        if joint_pos_end_effector is not None
        and joint_pos_limits_end_effector is not None
        else (
            joint_pos_end_effector
            if joint_pos_end_effector is not None
            else torch.empty((num_envs, 0), dtype=dtype, device=device)
        )
    )

    ## Kinematics
    fk_rotmat_end_effector = matrix_from_quat(fk_quat_end_effector)
    fk_rot6d_end_effector = rotmat_to_rot6d(fk_rotmat_end_effector)

    ## Transforms (world frame)
    # End-effector -> Object
    tf_pos_end_effector_to_objs, tf_quat_end_effector_to_objs = (
        subtract_frame_transforms(
            t01=tf_pos_end_effector.unsqueeze(1).repeat(1, num_objs, 1),
            q01=tf_quat_end_effector.unsqueeze(1).repeat(1, num_objs, 1),
            t02=tf_pos_objs,
            q02=tf_quat_objs,
        )
    )
    tf_rotmat_end_effector_to_objs = matrix_from_quat(tf_quat_end_effector_to_objs)
    tf_rot6d_end_effector_to_objs = rotmat_to_rot6d(tf_rotmat_end_effector_to_objs)

    # Peg -> Peg ends
    _tf_pos_pegs_end0, _ = combine_frame_transforms(
        t01=tf_pos_objs,
        q01=tf_quat_objs,
        t12=offset_pos_pegs_ends[:, :, 0],
    )
    _tf_pos_pegs_end1, _ = combine_frame_transforms(
        t01=tf_pos_objs,
        q01=tf_quat_objs,
        t12=offset_pos_pegs_ends[:, :, 1],
    )
    tf_pos_pegs_ends = torch.stack([_tf_pos_pegs_end0, _tf_pos_pegs_end1], dim=2)

    # Hole -> Hole entrance | bottom
    tf_pos_holes_entrance, _ = combine_frame_transforms(
        t01=tf_pos_targets,
        q01=tf_quat_targets,
        t12=offset_pos_holes_entrance,
    )
    tf_pos_holes_bottom, _ = combine_frame_transforms(
        t01=tf_pos_targets,
        q01=tf_quat_targets,
        t12=offset_pos_holes_bottom,
    )

    # Peg ends -> Hole entrance
    tf_pos_pegs_end0_to_holes_entrance, tf_quat_pegs_to_holes_entrance = (
        subtract_frame_transforms(
            t01=tf_pos_pegs_ends[:, :, 0],
            q01=tf_quat_objs,
            t02=tf_pos_holes_entrance,
            q02=tf_quat_targets,
        )
    )
    tf_pos_pegs_end1_to_holes_entrance, _ = subtract_frame_transforms(
        t01=tf_pos_pegs_ends[:, :, 1],
        q01=tf_quat_objs,
        t02=tf_pos_holes_entrance,
    )
    tf_pos_pegs_ends_to_holes_entrance = torch.stack(
        [tf_pos_pegs_end0_to_holes_entrance, tf_pos_pegs_end1_to_holes_entrance], dim=2
    )
    tf_rotmat_pegs_to_holes = matrix_from_quat(tf_quat_pegs_to_holes_entrance)
    tf_rot6d_pegs_to_holes = rotmat_to_rot6d(tf_rotmat_pegs_to_holes)

    # Peg ends -> Hole bottom
    tf_pos_pegs_end0_to_holes_bottom, _ = subtract_frame_transforms(
        t01=tf_pos_pegs_ends[:, :, 0],
        q01=tf_quat_objs,
        t02=tf_pos_holes_bottom,
    )
    tf_pos_pegs_end1_to_holes_bottom, _ = subtract_frame_transforms(
        t01=tf_pos_pegs_ends[:, :, 1],
        q01=tf_quat_objs,
        t02=tf_pos_holes_bottom,
    )
    tf_pos_pegs_ends_to_holes_bottom = torch.stack(
        [tf_pos_pegs_end0_to_holes_bottom, tf_pos_pegs_end1_to_holes_bottom], dim=1
    )

    # End-effector -> Panel CoM
    pos_robot_ee_to_panel, quat_robot_ee_to_panel = subtract_frame_transforms(
        t01=tf_pos_end_effector,
        q01=tf_quat_end_effector,
        t02=tf_pos_panel,
        q02=tf_quat_panel,
    )
    rotmat_robot_ee_to_panel = matrix_from_quat(quat_robot_ee_to_panel)
    rot6d_robot_ee_to_panel = rotmat_to_rot6d(rotmat_robot_ee_to_panel)

    # Panel CoM -> Panel Target
    pos_panel_to_panel_target, quat_panel_to_panel_target = subtract_frame_transforms(
        t01=tf_pos_panel,
        q01=tf_quat_panel,
        t02=tf_pos_panel_target,
        q02=tf_quat_panel_target,
    )
    rotmat_panel_to_panel_target = matrix_from_quat(quat_panel_to_panel_target)
    rot6d_panel_to_panel_target = rotmat_to_rot6d(rotmat_panel_to_panel_target)

    ## Contacts
    contact_forces_mean_robot = contact_forces_robot.mean(dim=1)
    contact_forces_mean_end_effector = (
        contact_forces_end_effector.mean(dim=1)
        if contact_forces_end_effector is not None
        else torch.empty((num_envs, 0), dtype=dtype, device=device)
    )
    contact_forces_end_effector = (
        contact_forces_end_effector
        if contact_forces_end_effector is not None
        else torch.empty((num_envs, 0), dtype=dtype, device=device)
    )

    #############
    ## Rewards ##
    #############
    # Penalty: Action rate
    WEIGHT_ACTION_RATE = -0.05
    penalty_action_rate = WEIGHT_ACTION_RATE * torch.sum(
        torch.square(act_current - act_previous), dim=1
    )

    # Penalty: Joint torque
    WEIGHT_JOINT_TORQUE = -0.000025
    MAX_JOINT_TORQUE_PENALTY = -4.0
    penalty_joint_torque = torch.clamp_min(
        WEIGHT_JOINT_TORQUE
        * torch.sum(torch.square(joint_applied_torque_robot), dim=1),
        min=MAX_JOINT_TORQUE_PENALTY,
    )

    # Penalty: Joint acceleration
    WEIGHT_JOINT_ACCELERATION = -0.0005
    MAX_JOINT_ACCELERATION_PENALTY = -4.0
    penalty_joint_acceleration = torch.clamp_min(
        WEIGHT_JOINT_ACCELERATION * torch.sum(torch.square(joint_acc_robot), dim=1),
        min=MAX_JOINT_ACCELERATION_PENALTY,
    )

    # Penalty: Undesired robot contacts
    WEIGHT_UNDESIRED_ROBOT_CONTACTS = -1.0
    THRESHOLD_UNDESIRED_ROBOT_CONTACTS = 10.0
    penalty_undesired_robot_contacts = WEIGHT_UNDESIRED_ROBOT_CONTACTS * (
        torch.max(torch.norm(contact_forces_robot, dim=-1), dim=1)[0]
        > THRESHOLD_UNDESIRED_ROBOT_CONTACTS
    )

    # Reward: End-effector top-down orientation
    WEIGHT_TOP_DOWN_ORIENTATION = 1.0
    TANH_STD_TOP_DOWN_ORIENTATION = 0.15
    top_down_alignment = torch.sum(
        fk_rotmat_end_effector[:, :, 2]
        * torch.tensor((0.0, 0.0, -1.0), device=device)
        .unsqueeze(0)
        .expand(num_envs, 3),
        dim=1,
    )
    reward_top_down_orientation = WEIGHT_TOP_DOWN_ORIENTATION * (
        1.0 - torch.tanh((1.0 - top_down_alignment) / TANH_STD_TOP_DOWN_ORIENTATION)
    )

    # Reward: Distance | End-effector <--> Object
    WEIGHT_DISTANCE_END_EFFECTOR_TO_OBJ = 2.5
    TANH_STD_DISTANCE_END_EFFECTOR_TO_OBJ = 0.2
    reward_distance_end_effector_to_objs = WEIGHT_DISTANCE_END_EFFECTOR_TO_OBJ * (
        1.0
        - torch.tanh(
            torch.norm(tf_pos_end_effector_to_objs, dim=-1)
            / TANH_STD_DISTANCE_END_EFFECTOR_TO_OBJ
        )
    ).sum(dim=-1)

    # Reward: Grasp object
    WEIGHT_GRASP = 8.0
    THRESHOLD_GRASP = 5.0
    reward_grasp = (
        WEIGHT_GRASP
        * (
            torch.mean(
                torch.max(
                    torch.norm(contact_force_matrix_end_effector, dim=-1), dim=-1
                )[0],
                dim=1,
            )
            > THRESHOLD_GRASP
        )
        if contact_force_matrix_end_effector is not None
        else torch.zeros(num_envs, dtype=dtype, device=device)
    )

    # Reward: Lift object
    WEIGHT_LIFT = 4.0
    HEIGHT_OFFSET_LIFT = 0.2
    HEIGHT_SPAN_LIFT = 0.1
    TANH_STD_HEIGHT_LIFT = 0.05
    reward_lift = WEIGHT_LIFT * (
        1.0
        - torch.tanh(
            (
                torch.abs(
                    tf_pos_objs[:, :, 2]
                    - tf_pos_objs_initial[:, :, 2]
                    - HEIGHT_OFFSET_LIFT
                )
                - HEIGHT_SPAN_LIFT
            ).clamp(min=0.0)
            / TANH_STD_HEIGHT_LIFT
        )
    ).sum(dim=-1)

    # Reward: Alignment | Peg -> Hole | Primary Z axis
    WEIGHT_ALIGN_PEG_TO_HOLE_PRIMARY = 8.0
    TANH_STD_ALIGN_PEG_TO_HOLE_PRIMARY = 0.5
    _pegs_to_holes_primary_axis_similarity = torch.abs(
        tf_rotmat_pegs_to_holes[:, :, 2, 2]
    )
    reward_align_pegs_to_holes_primary = WEIGHT_ALIGN_PEG_TO_HOLE_PRIMARY * (
        1.0
        - torch.tanh(
            (1.0 - _pegs_to_holes_primary_axis_similarity)
            / TANH_STD_ALIGN_PEG_TO_HOLE_PRIMARY
        )
    ).sum(dim=-1)

    # Reward: Alignment | Peg -> Hole | Secondary XY axes (affected by primary via power)
    WEIGHT_ALIGN_PEG_TO_HOLE_SECONDARY = 4.0
    TANH_STD_ALIGN_PEG_TO_HOLE_SECONDARY = 0.2
    _pegs_to_holes_yaw = torch.atan2(
        tf_rotmat_pegs_to_holes[:, :, 0, 1], tf_rotmat_pegs_to_holes[:, :, 0, 0]
    )
    _symmetry_step = 2 * torch.pi / symmetry_peg
    _pegs_to_holes_yaw_symmetric_directional = _pegs_to_holes_yaw % _symmetry_step
    # Note: Lines above might result in NaN/inf when `peg_rot_symmetry_n=0` (infinite circular symmetry)
    #       However, the following `torch.where()` will handle this case
    _pegs_to_holes_yaw_symmetric_normalized = torch.where(
        symmetry_peg <= 0,
        0.0,
        torch.min(
            _pegs_to_holes_yaw_symmetric_directional,
            _symmetry_step - _pegs_to_holes_yaw_symmetric_directional,
        )
        / (_symmetry_step / 2.0),
    )
    reward_align_pegs_to_holes_secondary = WEIGHT_ALIGN_PEG_TO_HOLE_SECONDARY * (
        1.0
        - torch.tanh(
            _pegs_to_holes_yaw_symmetric_normalized.pow(
                _pegs_to_holes_primary_axis_similarity
            )
            / TANH_STD_ALIGN_PEG_TO_HOLE_SECONDARY
        )
    ).sum(dim=-1)

    # Reward: Distance | Peg -> Hole entrance (gradual)
    WEIGHT_DISTANCE_PEG_TO_HOLE_ENTRANCE_GRADUAL = 8.0
    TANH_STD_DISTANCE_PEG_TO_HOLE_ENTRANCE_GRADUAL = 0.16
    reward_distance_pegs_to_holes_entrance_gradual = (
        WEIGHT_DISTANCE_PEG_TO_HOLE_ENTRANCE_GRADUAL
        * (
            1.0
            - torch.tanh(
                torch.min(
                    torch.norm(tf_pos_pegs_ends_to_holes_entrance, dim=-1), dim=1
                )[0]
                / TANH_STD_DISTANCE_PEG_TO_HOLE_ENTRANCE_GRADUAL
            )
        )
    ).sum(dim=-1)

    # Reward: Distance | Peg -> Hole entrance
    WEIGHT_DISTANCE_PEG_TO_HOLE_ENTRANCE = 32.0
    TANH_STD_DISTANCE_PEG_TO_HOLE_ENTRANCE = 0.04
    reward_distance_pegs_to_holes_entrance = WEIGHT_DISTANCE_PEG_TO_HOLE_ENTRANCE * (
        1.0
        - torch.tanh(
            torch.min(torch.norm(tf_pos_pegs_ends_to_holes_entrance, dim=-1), dim=1)[0]
            / TANH_STD_DISTANCE_PEG_TO_HOLE_ENTRANCE
        )
    ).sum(dim=-1)

    # Reward: Distance | Peg -> Hole bottom
    WEIGHT_DISTANCE_PEG_TO_HOLE_BOTTOM = 256.0
    TANH_STD_DISTANCE_PEG_TO_HOLE_BOTTOM = 0.002
    reward_distance_pegs_to_holes_bottom = WEIGHT_DISTANCE_PEG_TO_HOLE_BOTTOM * (
        1.0
        - torch.tanh(
            torch.min(torch.norm(tf_pos_pegs_ends_to_holes_bottom, dim=-1), dim=1)[0]
            / TANH_STD_DISTANCE_PEG_TO_HOLE_BOTTOM
        )
    ).sum(dim=-1)

    # Reward: Distance | End-effector <--> Panel
    WEIGHT_DISTANCE_EE_TO_PANEL = 4.0 * WEIGHT_DISTANCE_END_EFFECTOR_TO_OBJ
    TANH_STD_DISTANCE_EE_TO_PANEL = 0.25
    reward_distance_ee_to_panel = WEIGHT_DISTANCE_EE_TO_PANEL * (
        1.0
        - torch.tanh(
            torch.norm(pos_robot_ee_to_panel, dim=-1) / TANH_STD_DISTANCE_EE_TO_PANEL
        )
    )

    # Reward: Panel lift
    WEIGHT_PANEL_LIFT = 4.0 * WEIGHT_LIFT
    HEIGHT_OFFSET_PANEL_LIFT = 0.3
    HEIGHT_SPAN_PANEL_LIFT = 0.15
    TANH_STD_HEIGHT_PANEL_LIFT = 0.075
    panel_target_height_offset = (
        torch.abs(
            tf_pos_panel[:, 2] - tf_pos_panel_initial[:, 2] - HEIGHT_OFFSET_PANEL_LIFT
        )
        - HEIGHT_SPAN_PANEL_LIFT
    ).clamp(min=0.0)
    reward_panel_lift = WEIGHT_PANEL_LIFT * (
        1.0 - torch.tanh(panel_target_height_offset / TANH_STD_HEIGHT_PANEL_LIFT)
    )

    # Reward: Distance | Panel <--> Panel Target (gra`dual)
    WEIGHT_DISTANCE_PANEL_TO_TARGET_GRADUAL = (
        4.0 * WEIGHT_DISTANCE_PEG_TO_HOLE_ENTRANCE_GRADUAL
    )
    TANH_STD_DISTANCE_PANEL_TO_TARGET_GRADUAL = 0.1
    reward_distance_panel_to_target_gradual = (
        WEIGHT_DISTANCE_PANEL_TO_TARGET_GRADUAL
        * (
            1.0
            - torch.tanh(
                torch.norm(pos_panel_to_panel_target, dim=-1)
                / TANH_STD_DISTANCE_PANEL_TO_TARGET_GRADUAL
            )
        )
    )

    # Reward: Distance | Panel <--> Panel Target
    WEIGHT_DISTANCE_PANEL_TO_TARGET = 4.0 * WEIGHT_DISTANCE_PEG_TO_HOLE_BOTTOM
    TANH_STD_DISTANCE_PANEL_TO_TARGET = 0.01
    reward_distance_panel_to_target = WEIGHT_DISTANCE_PANEL_TO_TARGET * (
        1.0
        - torch.tanh(
            torch.norm(pos_panel_to_panel_target, dim=-1)
            / TANH_STD_DISTANCE_PANEL_TO_TARGET
        )
    )

    ##################
    ## Terminations ##
    ##################
    # No termination condition
    termination = torch.zeros(num_envs, dtype=torch.bool, device=device)
    # Truncation
    truncation = (
        episode_length >= max_episode_length
        if truncate_episodes
        else torch.zeros(num_envs, dtype=torch.bool, device=device)
    )

    return StepReturn(
        {
            "state": {
                "contact_forces_mean_robot": contact_forces_mean_robot,
                "contact_forces_mean_end_effector": contact_forces_mean_end_effector,
                "tf_pos_end_effector_to_objs": tf_pos_end_effector_to_objs,
                "tf_rot6d_end_effector_to_objs": tf_rot6d_end_effector_to_objs,
                "tf_pos_pegs_end0_to_holes_entrance": tf_pos_pegs_end0_to_holes_entrance,
                "tf_pos_pegs_end1_to_holes_entrance": tf_pos_pegs_end1_to_holes_entrance,
                "tf_pos_pegs_end0_to_holes_bottom": tf_pos_pegs_end0_to_holes_bottom,
                "tf_pos_pegs_end1_to_holes_bottom": tf_pos_pegs_end1_to_holes_bottom,
                "tf_rot6d_pegs_to_holes": tf_rot6d_pegs_to_holes,
                "pos_robot_ee_to_panel": pos_robot_ee_to_panel,
                "rot6d_robot_ee_to_panel": rot6d_robot_ee_to_panel,
                "pos_panel_to_panel_target": pos_panel_to_panel_target,
                "rot6d_panel_to_panel_target": rot6d_panel_to_panel_target,
            },
            "state_dyn": {
                "contact_forces_robot": contact_forces_robot,
                "contact_forces_end_effector": contact_forces_end_effector,
            },
            "proprio": {
                "fk_pos_end_effector": fk_pos_end_effector,
                "fk_rot6d_end_effector": fk_rot6d_end_effector,
            },
            "proprio_dyn": {
                "joint_pos_robot_normalized": joint_pos_robot_normalized,
                "joint_pos_end_effector_normalized": joint_pos_end_effector_normalized,
                "joint_acc_robot": joint_acc_robot,
                "joint_applied_torque_robot": joint_applied_torque_robot,
            },
        },
        {
            "penalty_action_rate": penalty_action_rate,
            "penalty_joint_torque": penalty_joint_torque,
            "penalty_joint_acceleration": penalty_joint_acceleration,
            "penalty_undesired_robot_contacts": penalty_undesired_robot_contacts,
            "reward_top_down_orientation": reward_top_down_orientation,
            "reward_distance_end_effector_to_objs": reward_distance_end_effector_to_objs,
            "reward_grasp": reward_grasp,
            "reward_lift": reward_lift,
            "reward_align_pegs_to_holes_primary": reward_align_pegs_to_holes_primary,
            "reward_align_pegs_to_holes_secondary": reward_align_pegs_to_holes_secondary,
            "reward_distance_pegs_to_holes_entrance_gradual": reward_distance_pegs_to_holes_entrance_gradual,
            "reward_distance_pegs_to_holes_entrance": reward_distance_pegs_to_holes_entrance,
            "reward_distance_pegs_to_holes_bottom": reward_distance_pegs_to_holes_bottom,
            "reward_distance_ee_to_panel": reward_distance_ee_to_panel,
            "reward_panel_lift": reward_panel_lift,
            "reward_distance_panel_to_target_gradual": reward_distance_panel_to_target_gradual,
            "reward_distance_panel_to_target": reward_distance_panel_to_target,
        },
        termination,
        truncation,
    )



from .env import MobileEnv, MobileEnvCfg, MobileEventCfg, MobileSceneCfg  # noqa: F401

# isort: split

from .aerial import *  # noqa: F403
from .ground import *  # noqa: F403
from .orbital import *  # noqa: F403



from typing import TYPE_CHECKING, Tuple

from pydantic import BaseModel

from srb import assets
from srb.core.asset import RigidObjectCfg

if TYPE_CHECKING:
    from srb._typing import AnyEnvCfg


class PanelCfg(BaseModel, arbitrary_types_allowed=True):
    asset_cfg: RigidObjectCfg
    offset_pos: Tuple[float, float, float] = (0.0, 0.0, 0.15)


@staticmethod
def select_solar_panel(
    env_cfg: "AnyEnvCfg", *, init_state: RigidObjectCfg.InitialStateCfg
) -> PanelCfg:
    offset_pos = (0.0, 0.0, 0.15)
    init_state.pos = (
        init_state.pos[0] + offset_pos[0],
        init_state.pos[1] + offset_pos[1],
        init_state.pos[2] + offset_pos[2],
    )

    cfg = assets.SolarPanel().asset_cfg
    cfg.init_state = init_state

    return PanelCfg(
        asset_cfg=cfg,
        offset_pos=offset_pos,
    )



from srb.core.action import (
    ActionGroup,
    BinaryJointPositionActionCfg,
    JointPositionBinaryActionGroup,
)
from srb.core.actuator import ImplicitActuatorCfg
from srb.core.asset import ActiveTool, ArticulationCfg, Frame, Transform
from srb.core.sim import (
    ArticulationRootPropertiesCfg,
    CollisionPropertiesCfg,
    RigidBodyPropertiesCfg,
    UsdFileCfg,
)
from srb.utils.math import deg_to_rad
from srb.utils.path import SRB_ASSETS_DIR_SRB_ROBOT


class SOArm100Gripper(ActiveTool):
    ## Model
    asset_cfg: ArticulationCfg = ArticulationCfg(
        prim_path="{ENV_REGEX_NS}/so_arm100_gripper",
        spawn=UsdFileCfg(
            usd_path=SRB_ASSETS_DIR_SRB_ROBOT.joinpath("gripper")
            .joinpath("so_arm100_5dof_gripper.usdz")
            .as_posix(),
            activate_contact_sensors=True,
            collision_props=CollisionPropertiesCfg(
                contact_offset=0.005, rest_offset=0.0
            ),
            rigid_props=RigidBodyPropertiesCfg(
                disable_gravity=True,
                max_depenetration_velocity=5.0,
            ),
            articulation_props=ArticulationRootPropertiesCfg(
                enabled_self_collisions=False,
                solver_position_iteration_count=8,
                solver_velocity_iteration_count=0,
            ),
        ),
        init_state=ArticulationCfg.InitialStateCfg(
            joint_pos={"Gripper": 0.0},
        ),
        actuators={
            "gripper": ImplicitActuatorCfg(
                joint_names_expr=["Gripper"],
                velocity_limit=1.5,
                effort_limit=2.0,
                stiffness=10.0,
                damping=1.0,
            ),
        },
    )

    ## Actions
    actions: ActionGroup = JointPositionBinaryActionGroup(
        BinaryJointPositionActionCfg(
            asset_name="robot",
            joint_names=["Gripper"],
            close_command_expr={"Gripper": 0.0},
            open_command_expr={"Gripper": deg_to_rad(60.0)},
        ),
    )

    ## Frames
    frame_mount: Frame = Frame(prim_relpath="Fixed_Gripper")
    frame_tool_centre_point: Frame = Frame(
        prim_relpath="Fixed_Gripper",
        offset=Transform(pos=(0.0, 0.0, 0.095)),
    )



from __future__ import annotations

from typing import Sequence, Type

from srb.core.asset.common import Frame
from srb.core.asset.robot.mobile.mobile_robot import MobileRobot, MobileRobotRegistry
from srb.core.asset.robot.mobile.mobile_robot_type import MobileRobotType


class AerialRobot(MobileRobot, mobile_robot_entrypoint=MobileRobotType.AERIAL):
    ## Frames
    frame_onboard_camera: Frame

    @classmethod
    def mobile_robot_registry(cls) -> Sequence[Type[AerialRobot]]:
        return MobileRobotRegistry.registry.get(MobileRobotType.AERIAL, [])  # type: ignore


class Multicopter(
    AerialRobot, mobile_robot_metaclass=True, arbitrary_types_allowed=True
):
    pass



from .manipulation import *  # noqa: F403
from .mobile import *  # noqa: F403
from .mobile_manipulation import *  # noqa: F403
from .robot import Robot, RobotRegistry  # noqa: F401
from .robot_type import RobotType  # noqa: F401



from dataclasses import MISSING
from typing import Sequence

import torch

from srb._typing import StepReturn
from srb.core.asset import (
    Articulation,
    RigidObjectCfg,
    RigidObjectCollection,
    RigidObjectCollectionCfg,
)
from srb.core.env import ManipulationEnv
from srb.core.manager import EventTermCfg, SceneEntityCfg
from srb.core.marker import VisualizationMarkers
from srb.core.mdp import reset_collection_root_state_uniform_poisson_disk_2d
from srb.core.sensor import ContactSensor, ContactSensorCfg
from srb.utils.cfg import configclass
from srb.utils.math import (
    matrix_from_quat,
    rotmat_to_rot6d,
    scale_transform,
    subtract_frame_transforms,
)

from .asset import select_sample
from .task import EventCfg, SceneCfg, TaskCfg

##############
### Config ###
##############


@configclass
class MultiSceneCfg(SceneCfg):
    sample: None = None
    samples: RigidObjectCollectionCfg = RigidObjectCollectionCfg(
        rigid_objects=MISSING,  # type: ignore
    )


@configclass
class MultiEventCfg(EventCfg):
    randomize_object_state: EventTermCfg = EventTermCfg(
        func=reset_collection_root_state_uniform_poisson_disk_2d,
        mode="reset",
        params={
            "asset_cfg": SceneEntityCfg("samples"),
            "pose_range": MISSING,
            "velocity_range": MISSING,
            "radius": 0.1,
        },
    )


@configclass
class MultiTaskCfg(TaskCfg):
    ## Scene
    scene: MultiSceneCfg = MultiSceneCfg()
    num_samples: int = 3

    ## Events
    events: MultiEventCfg = MultiEventCfg()

    ## Time
    episode_length_s: float = MISSING  # type: ignore
    _base_episode_length_s: float = 7.5

    def __post_init__(self):
        super().__post_init__()

        # Time
        self.episode_length_s = self.num_samples * self._base_episode_length_s

        # Scene: Samples
        self.scene.sample = None
        samples = [
            select_sample(
                self,
                seed=self.seed + (i * self.scene.num_envs),
                prim_path=f"{{ENV_REGEX_NS}}/sample{i}",
                asset_cfg=SceneEntityCfg(f"sample{i}"),
                init_state=RigidObjectCfg.InitialStateCfg(pos=(0.55, 0.0, 0.0)),
                activate_contact_sensors=True,
            )
            for i in range(self.num_samples)
        ]
        self.scene.samples.rigid_objects = {
            f"sample{i}": cfg.asset_cfg for i, cfg in enumerate(samples)
        }

        # Sensor: End-effector contacts
        if isinstance(self.scene.contacts_end_effector, ContactSensorCfg):
            self.scene.contacts_end_effector.filter_prim_paths_expr = [
                f"{{ENV_REGEX_NS}}/sample{i}" for i in range(self.num_samples)
            ]

        # Events: Randomize object state
        self.events.randomize_object_state.params["pose_range"] = samples[
            0
        ].state_randomizer.params["pose_range"]
        self.events.randomize_object_state.params["velocity_range"] = samples[
            0
        ].state_randomizer.params["velocity_range"]

        # Update seed & number of variants for procedural assets
        self._update_procedural_assets()


############
### Task ###
############


class MultiTask(ManipulationEnv):
    cfg: MultiTaskCfg

    def __init__(self, cfg: MultiTaskCfg, **kwargs):
        super().__init__(cfg, **kwargs)

        ## Get scene assets
        self._objs: RigidObjectCollection = self.scene["samples"]
        self._target_marker: VisualizationMarkers = VisualizationMarkers(
            self.cfg.target_marker_cfg
        )

        ## Initialize buffers
        self._tf_pos_objs_initial = torch.zeros(
            (self.num_envs, self.cfg.num_samples, 3),
            dtype=torch.float32,
            device=self.device,
        )
        self._tf_pos_target = self.scene.env_origins.unsqueeze(1) + torch.tensor(
            self.cfg.tf_pos_target, dtype=torch.float32, device=self.device
        ).repeat(self.num_envs, self.cfg.num_samples, 1)
        self._tf_quat_target = torch.tensor(
            self.cfg.tf_quat_target, dtype=torch.float32, device=self.device
        ).repeat(self.num_envs, self.cfg.num_samples, 1)

        ## Visualize target
        self._target_marker.visualize(
            self._tf_pos_target[:, 0, :], self._tf_quat_target[:, 0, :]
        )

    def _reset_idx(self, env_ids: Sequence[int]):
        super()._reset_idx(env_ids)
        self._tf_pos_objs_initial[env_ids] = self._objs.data.object_com_pos_w[env_ids]

    def extract_step_return(self) -> StepReturn:
        return _compute_step_return(
            ## Time
            episode_length=self.episode_length_buf,
            max_episode_length=self.max_episode_length,
            truncate_episodes=self.cfg.truncate_episodes,
            ## Actions
            act_current=self.action_manager.action,
            act_previous=self.action_manager.prev_action,
            ## States
            # Joints
            joint_pos_robot=self._robot.data.joint_pos,
            joint_pos_limits_robot=(
                self._robot.data.soft_joint_pos_limits
                if torch.all(torch.isfinite(self._robot.data.soft_joint_pos_limits))
                else None
            ),
            joint_pos_end_effector=self._end_effector.data.joint_pos
            if isinstance(self._end_effector, Articulation)
            else None,
            joint_pos_limits_end_effector=(
                self._end_effector.data.soft_joint_pos_limits
                if isinstance(self._end_effector, Articulation)
                and torch.all(
                    torch.isfinite(self._end_effector.data.soft_joint_pos_limits)
                )
                else None
            ),
            joint_acc_robot=self._robot.data.joint_acc,
            joint_applied_torque_robot=self._robot.data.applied_torque,
            # Kinematics
            fk_pos_end_effector=self._tf_end_effector.data.target_pos_source[:, 0, :],
            fk_quat_end_effector=self._tf_end_effector.data.target_quat_source[:, 0, :],
            # Transforms (world frame)
            tf_pos_end_effector=self._tf_end_effector.data.target_pos_w[:, 0, :],
            tf_quat_end_effector=self._tf_end_effector.data.target_quat_w[:, 0, :],
            tf_pos_objs_initial=self._tf_pos_objs_initial,
            tf_pos_objs=self._objs.data.object_com_pos_w,
            tf_quat_objs=self._objs.data.object_com_quat_w,
            tf_pos_target=self._tf_pos_target,
            tf_quat_target=self._tf_quat_target,
            # Contacts
            contact_forces_robot=self._contacts_robot.data.net_forces_w,  # type: ignore
            contact_forces_end_effector=self._contacts_end_effector.data.net_forces_w
            if isinstance(self._contacts_end_effector, ContactSensor)
            else None,
            contact_force_matrix_end_effector=self._contacts_end_effector.data.force_matrix_w
            if isinstance(self._contacts_end_effector, ContactSensor)
            else None,
        )


@torch.jit.script
def _compute_step_return(
    *,
    ## Time
    episode_length: torch.Tensor,
    max_episode_length: int,
    truncate_episodes: bool,
    ## Actions
    act_current: torch.Tensor,
    act_previous: torch.Tensor,
    ## States
    # Joints
    joint_pos_robot: torch.Tensor,
    joint_pos_limits_robot: torch.Tensor | None,
    joint_pos_end_effector: torch.Tensor | None,
    joint_pos_limits_end_effector: torch.Tensor | None,
    joint_acc_robot: torch.Tensor,
    joint_applied_torque_robot: torch.Tensor,
    # Kinematics
    fk_pos_end_effector: torch.Tensor,
    fk_quat_end_effector: torch.Tensor,
    # Transforms (world frame)
    tf_pos_end_effector: torch.Tensor,
    tf_quat_end_effector: torch.Tensor,
    tf_pos_objs_initial: torch.Tensor,
    tf_pos_objs: torch.Tensor,
    tf_quat_objs: torch.Tensor,
    tf_pos_target: torch.Tensor,
    tf_quat_target: torch.Tensor,
    # Contacts
    contact_forces_robot: torch.Tensor,
    contact_forces_end_effector: torch.Tensor | None,
    contact_force_matrix_end_effector: torch.Tensor | None,
) -> StepReturn:
    num_envs = episode_length.size(0)
    num_objs = tf_pos_objs.size(1)
    dtype = episode_length.dtype
    device = episode_length.device

    ############
    ## States ##
    ############
    ## Joints
    # Robot joints
    joint_pos_robot_normalized = (
        scale_transform(
            joint_pos_robot,
            joint_pos_limits_robot[:, :, 0],
            joint_pos_limits_robot[:, :, 1],
        )
        if joint_pos_limits_robot is not None
        else joint_pos_robot
    )
    # End-effector joints
    joint_pos_end_effector_normalized = (
        scale_transform(
            joint_pos_end_effector,
            joint_pos_limits_end_effector[:, :, 0],
            joint_pos_limits_end_effector[:, :, 1],
        )
        if joint_pos_end_effector is not None
        and joint_pos_limits_end_effector is not None
        else (
            joint_pos_end_effector
            if joint_pos_end_effector is not None
            else torch.empty((num_envs, 0), dtype=dtype, device=device)
        )
    )

    ## Kinematics
    fk_rotmat_end_effector = matrix_from_quat(fk_quat_end_effector)
    fk_rot6d_end_effector = rotmat_to_rot6d(fk_rotmat_end_effector)

    ## Transforms (world frame)
    # End-effector -> Object
    tf_pos_end_effector_to_objs, tf_quat_end_effector_to_objs = (
        subtract_frame_transforms(
            t01=tf_pos_end_effector.unsqueeze(1).repeat(1, num_objs, 1),
            q01=tf_quat_end_effector.unsqueeze(1).repeat(1, num_objs, 1),
            t02=tf_pos_objs,
            q02=tf_quat_objs,
        )
    )
    tf_rotmat_end_effector_to_objs = matrix_from_quat(tf_quat_end_effector_to_objs)
    tf_rot6d_end_effector_to_objs = rotmat_to_rot6d(tf_rotmat_end_effector_to_objs)
    # Object -> Target
    tf_pos_objs_to_target, tf_quat_objs_to_target = subtract_frame_transforms(
        t01=tf_pos_objs,
        q01=tf_quat_objs,
        t02=tf_pos_target,
        q02=tf_quat_target,
    )
    tf_rotmat_objs_to_target = matrix_from_quat(tf_quat_objs_to_target)
    tf_rot6d_objs_to_target = rotmat_to_rot6d(tf_rotmat_objs_to_target)

    ## Contacts
    contact_forces_mean_robot = contact_forces_robot.mean(dim=1)
    contact_forces_mean_end_effector = (
        contact_forces_end_effector.mean(dim=1)
        if contact_forces_end_effector is not None
        else torch.empty((num_envs, 0), dtype=dtype, device=device)
    )
    contact_forces_end_effector = (
        contact_forces_end_effector
        if contact_forces_end_effector is not None
        else torch.empty((num_envs, 0), dtype=dtype, device=device)
    )

    #############
    ## Rewards ##
    #############
    # Penalty: Action rate
    WEIGHT_ACTION_RATE = -0.05
    penalty_action_rate = WEIGHT_ACTION_RATE * torch.sum(
        torch.square(act_current - act_previous), dim=1
    )

    # Penalty: Joint torque
    WEIGHT_JOINT_TORQUE = -0.000025
    MAX_JOINT_TORQUE_PENALTY = -4.0
    penalty_joint_torque = torch.clamp_min(
        WEIGHT_JOINT_TORQUE
        * torch.sum(torch.square(joint_applied_torque_robot), dim=1),
        min=MAX_JOINT_TORQUE_PENALTY,
    )

    # Penalty: Joint acceleration
    WEIGHT_JOINT_ACCELERATION = -0.0005
    MAX_JOINT_ACCELERATION_PENALTY = -4.0
    penalty_joint_acceleration = torch.clamp_min(
        WEIGHT_JOINT_ACCELERATION * torch.sum(torch.square(joint_acc_robot), dim=1),
        min=MAX_JOINT_ACCELERATION_PENALTY,
    )

    # Penalty: Undesired robot contacts
    WEIGHT_UNDESIRED_ROBOT_CONTACTS = -1.0
    THRESHOLD_UNDESIRED_ROBOT_CONTACTS = 10.0
    penalty_undesired_robot_contacts = WEIGHT_UNDESIRED_ROBOT_CONTACTS * (
        torch.max(torch.norm(contact_forces_robot, dim=-1), dim=1)[0]
        > THRESHOLD_UNDESIRED_ROBOT_CONTACTS
    )

    # Reward: End-effector top-down orientation
    WEIGHT_TOP_DOWN_ORIENTATION = 1.0
    TANH_STD_TOP_DOWN_ORIENTATION = 0.15
    top_down_alignment = torch.sum(
        fk_rotmat_end_effector[:, :, 2]
        * torch.tensor((0.0, 0.0, -1.0), device=device)
        .unsqueeze(0)
        .expand(num_envs, 3),
        dim=1,
    )
    reward_top_down_orientation = WEIGHT_TOP_DOWN_ORIENTATION * (
        1.0 - torch.tanh((1.0 - top_down_alignment) / TANH_STD_TOP_DOWN_ORIENTATION)
    )

    # Reward: Distance | End-effector <--> Object
    WEIGHT_DISTANCE_END_EFFECTOR_TO_OBJ = 2.5
    TANH_STD_DISTANCE_END_EFFECTOR_TO_OBJ = 0.2
    reward_distance_end_effector_to_objs = WEIGHT_DISTANCE_END_EFFECTOR_TO_OBJ * (
        1.0
        - torch.tanh(
            torch.norm(tf_pos_end_effector_to_objs, dim=-1)
            / TANH_STD_DISTANCE_END_EFFECTOR_TO_OBJ
        )
    ).sum(dim=1)

    # Reward: Grasp object
    WEIGHT_GRASP = 4.0
    THRESHOLD_GRASP = 5.0
    reward_grasp = (
        WEIGHT_GRASP
        * (
            torch.mean(
                torch.max(
                    torch.norm(contact_force_matrix_end_effector, dim=-1), dim=-1
                )[0],
                dim=1,
            )
            > THRESHOLD_GRASP
        )
        if contact_force_matrix_end_effector is not None
        else torch.zeros(num_envs, dtype=dtype, device=device)
    )

    # Reward: Lift object
    WEIGHT_LIFT = 8.0
    HEIGHT_OFFSET_LIFT = 0.5
    HEIGHT_SPAN_LIFT = 0.25
    TANH_STD_HEIGHT_LIFT = 0.1
    reward_lift = WEIGHT_LIFT * (
        1.0
        - torch.tanh(
            (
                torch.abs(
                    tf_pos_objs[:, :, 2]
                    - tf_pos_objs_initial[:, :, 2]
                    - HEIGHT_OFFSET_LIFT
                )
                - HEIGHT_SPAN_LIFT
            ).clamp(min=0.0)
            / TANH_STD_HEIGHT_LIFT
        )
    ).sum(dim=1)

    # Reward: Distance | Object <--> Target
    WEIGHT_DISTANCE_OBJ_TO_TARGET = 32.0
    TANH_STD_DISTANCE_OBJ_TO_TARGET = 0.2
    reward_distance_objs_to_target = WEIGHT_DISTANCE_OBJ_TO_TARGET * (
        1.0
        - torch.tanh(
            torch.norm(tf_pos_objs_to_target, dim=-1) / TANH_STD_DISTANCE_OBJ_TO_TARGET
        )
    ).sum(dim=1)

    ##################
    ## Terminations ##
    ##################
    # No termination condition
    termination = torch.zeros(num_envs, dtype=torch.bool, device=device)
    # Truncation
    truncation = (
        episode_length >= max_episode_length
        if truncate_episodes
        else torch.zeros(num_envs, dtype=torch.bool, device=device)
    )

    return StepReturn(
        {
            "state": {
                "contact_forces_mean_robot": contact_forces_mean_robot,
                "contact_forces_mean_end_effector": contact_forces_mean_end_effector,
                "tf_pos_end_effector_to_objs": tf_pos_end_effector_to_objs,
                "tf_rot6d_end_effector_to_objs": tf_rot6d_end_effector_to_objs,
                "tf_pos_objs_to_target": tf_pos_objs_to_target,
                "tf_rot6d_objs_to_target": tf_rot6d_objs_to_target,
            },
            "state_dyn": {
                "contact_forces_robot": contact_forces_robot,
                "contact_forces_end_effector": contact_forces_end_effector,
            },
            "proprio": {
                "fk_pos_end_effector": fk_pos_end_effector,
                "fk_rot6d_end_effector": fk_rot6d_end_effector,
            },
            "proprio_dyn": {
                "joint_pos_robot_normalized": joint_pos_robot_normalized,
                "joint_pos_end_effector_normalized": joint_pos_end_effector_normalized,
                "joint_acc_robot": joint_acc_robot,
                "joint_applied_torque_robot": joint_applied_torque_robot,
            },
        },
        {
            "penalty_action_rate": penalty_action_rate,
            "penalty_joint_torque": penalty_joint_torque,
            "penalty_joint_acceleration": penalty_joint_acceleration,
            "penalty_undesired_robot_contacts": penalty_undesired_robot_contacts,
            "reward_top_down_orientation": reward_top_down_orientation,
            "reward_distance_end_effector_to_objs": reward_distance_end_effector_to_objs,
            "reward_grasp": reward_grasp,
            "reward_lift": reward_lift,
            "reward_distance_objs_to_target": reward_distance_objs_to_target,
        },
        termination,
        truncation,
    )



from .multicopter_body_acc import (  # noqa: F401
    MulticopterBodyAccelerationAction,
    MulticopterBodyAccelerationActionCfg,
)
from .thrust import ThrustAction, ThrustActionCfg, ThrusterCfg  # noqa: F401
from .wheeled_drive import WheeledDriveAction, WheeledDriveActionCfg  # noqa: F401



from typing import Sequence, Tuple

import torch

from srb import assets
from srb._typing import StepReturn
from srb.core.action import ThrustAction
from srb.core.asset import RigidObject, RigidObjectCfg
from srb.core.env import OrbitalEnv, OrbitalEnvCfg, OrbitalEventCfg, OrbitalSceneCfg
from srb.core.manager import EventTermCfg, SceneEntityCfg
from srb.core.marker import VisualizationMarkers, VisualizationMarkersCfg
from srb.core.mdp import reset_root_state_uniform
from srb.core.sim import ArrowCfg, PreviewSurfaceCfg
from srb.utils.cfg import configclass
from srb.utils.math import (
    combine_frame_transforms,
    deg_to_rad,
    matrix_from_quat,
    rotmat_to_rot6d,
    rpy_to_quat,
    subtract_frame_transforms,
)

##############
### Config ###
##############


@configclass
class SceneCfg(OrbitalSceneCfg):
    ## Assets
    target: RigidObjectCfg = assets.Cubesat().asset_cfg
    target.prim_path = "{ENV_REGEX_NS}/target"


@configclass
class EventCfg(OrbitalEventCfg):
    randomize_target_state: EventTermCfg = EventTermCfg(
        func=reset_root_state_uniform,
        mode="reset",
        params={
            "asset_cfg": SceneEntityCfg("target"),
            "pose_range": {
                "x": (0.5, 2.0),
                "y": (-2.0, 2.0),
                "z": (-2.0, 2.0),
                "roll": (-torch.pi, torch.pi),
                "pitch": (-torch.pi, torch.pi),
                "yaw": (-torch.pi, torch.pi),
            },
            "velocity_range": {
                "x": (-0.05, 0.05),
                "y": (-0.05, 0.05),
                "z": (-0.05, 0.05),
                "roll": (-deg_to_rad(10.0), deg_to_rad(10.0)),
                "pitch": (-deg_to_rad(10.0), deg_to_rad(10.0)),
                "yaw": (-deg_to_rad(10.0), deg_to_rad(10.0)),
            },
        },
    )


@configclass
class TaskCfg(OrbitalEnvCfg):
    ## Scene
    scene: SceneCfg = SceneCfg()

    ## Events
    events: EventCfg = EventCfg()

    ## Time
    episode_length_s: float = 25.0
    is_finite_horizon: bool = True

    ## Target offset
    target_offset_pos: Tuple[float, float, float] = (0.0, 0.0, 0.5)
    target_offset_quat: Tuple[float, float, float, float] = rpy_to_quat(0.0, 90.0, 0.0)
    target_marker_cfg: VisualizationMarkersCfg = VisualizationMarkersCfg(
        prim_path="/Visuals/target",
        markers={
            "target": ArrowCfg(
                tail_radius=0.01,
                tail_length=0.05,
                head_radius=0.02,
                head_length=0.025,
                visual_material=PreviewSurfaceCfg(emissive_color=(0.2, 0.2, 0.8)),
            )
        },
    )

    def __post_init__(self):
        super().__post_init__()

        # Scene: Target
        if hasattr(self.scene.target.spawn, "seed"):
            self.scene.target.spawn.seed = self.seed + self.scene.num_envs  # type: ignore


############
### Task ###
############


class Task(OrbitalEnv):
    cfg: TaskCfg

    def __init__(self, cfg: TaskCfg, **kwargs):
        super().__init__(cfg, **kwargs)

        ## Get scene assets
        self._target: RigidObject = self.scene["target"]
        self._target_marker: VisualizationMarkers = VisualizationMarkers(
            self.cfg.target_marker_cfg
        )

        ## Initialize buffers
        self._target_offset_pos = torch.tensor(
            self.cfg.target_offset_pos, dtype=torch.float32, device=self.device
        ).repeat(self.num_envs, 1)
        self._target_offset_quat = torch.tensor(
            self.cfg.target_offset_quat, dtype=torch.float32, device=self.device
        ).repeat(self.num_envs, 1)

    def _reset_idx(self, env_ids: Sequence[int]):
        super()._reset_idx(env_ids)

    def extract_step_return(self) -> StepReturn:
        ## Compute the target pose with offset
        tf_pos_target, tf_quat_target = combine_frame_transforms(
            self._target.data.root_pos_w,
            self._target.data.root_quat_w,
            self._target_offset_pos,
            self._target_offset_quat,
        )

        ## Visualize target
        self._target_marker.visualize(tf_pos_target, tf_quat_target)

        ## Get remaining fuel (if applicable)
        if self._thrust_action_term_key:
            thrust_action_term: ThrustAction = self.action_manager._terms[  # type: ignore
                self._thrust_action_term_key
            ]
            remaining_fuel = (
                thrust_action_term.remaining_fuel / thrust_action_term.cfg.fuel_capacity
            ).unsqueeze(-1)
        else:
            remaining_fuel = None

        return _compute_step_return(
            ## Time
            episode_length=self.episode_length_buf,
            max_episode_length=self.max_episode_length,
            truncate_episodes=self.cfg.truncate_episodes,
            ## Actions
            act_current=self.action_manager.action,
            act_previous=self.action_manager.prev_action,
            ## States
            # Root
            tf_pos_robot=self._robot.data.root_pos_w,
            tf_quat_robot=self._robot.data.root_quat_w,
            vel_lin_robot=self._robot.data.root_lin_vel_b,
            vel_ang_robot=self._robot.data.root_ang_vel_b,
            vel_lin_target=self._target.data.root_lin_vel_b,
            vel_ang_target=self._target.data.root_ang_vel_b,
            # Transforms (world frame)
            tf_pos_target=tf_pos_target,
            tf_quat_target=tf_quat_target,
            # IMU
            imu_lin_acc=self._imu_robot.data.lin_acc_b,
            imu_ang_vel=self._imu_robot.data.ang_vel_b,
            # Fuel
            remaining_fuel=remaining_fuel,
        )


@torch.jit.script
def _compute_step_return(
    *,
    ## Time
    episode_length: torch.Tensor,
    max_episode_length: int,
    truncate_episodes: bool,
    ## Actions
    act_current: torch.Tensor,
    act_previous: torch.Tensor,
    ## States
    # Root
    tf_pos_robot: torch.Tensor,
    tf_quat_robot: torch.Tensor,
    vel_lin_robot: torch.Tensor,
    vel_ang_robot: torch.Tensor,
    vel_lin_target: torch.Tensor,
    vel_ang_target: torch.Tensor,
    # Transforms (world frame)
    tf_pos_target: torch.Tensor,
    tf_quat_target: torch.Tensor,
    # IMU
    imu_lin_acc: torch.Tensor,
    imu_ang_vel: torch.Tensor,
    # Fuel
    remaining_fuel: torch.Tensor | None,
) -> StepReturn:
    num_envs = episode_length.size(0)
    dtype = episode_length.dtype
    device = episode_length.device

    ############
    ## States ##
    ############
    ## Root
    tf_rotmat_robot = matrix_from_quat(tf_quat_robot)
    tf_rot6d_robot = rotmat_to_rot6d(tf_rotmat_robot)

    ## Transforms (world frame)
    # Robot -> Target
    tf_pos_robot_to_target, tf_quat_robot_to_target = subtract_frame_transforms(
        t01=tf_pos_robot,
        q01=tf_quat_robot,
        t02=tf_pos_target,
        q02=tf_quat_target,
    )
    tf_rotmat_robot_to_target = matrix_from_quat(tf_quat_robot_to_target)
    tf_rot6d_robot_to_target = rotmat_to_rot6d(tf_rotmat_robot_to_target)

    distance_robot_to_target = torch.norm(tf_pos_robot_to_target, dim=-1)

    ## Fuel
    remaining_fuel = (
        remaining_fuel
        if remaining_fuel is not None
        else torch.ones((num_envs, 1), dtype=dtype, device=device)
    )

    #############
    ## Rewards ##
    #############
    # Penalty: Action rate
    WEIGHT_ACTION_RATE = -0.05
    penalty_action_rate = WEIGHT_ACTION_RATE * torch.sum(
        torch.square(act_current - act_previous), dim=1
    )

    # Penalty: Fuel consumption
    WEIGHT_FUEL_CONSUMPTION = -8.0
    penalty_fuel_consumption = WEIGHT_FUEL_CONSUMPTION * torch.square(
        1.0 - remaining_fuel.squeeze(-1)
    )

    # Penalty: Angular velocity
    WEIGHT_ANGULAR_VELOCITY = -0.05
    MAX_ANGULAR_VELOCITY_PENALTY = -5.0
    penalty_angular_velocity = torch.clamp_min(
        WEIGHT_ANGULAR_VELOCITY * torch.sum(torch.square(vel_ang_robot), dim=1),
        min=MAX_ANGULAR_VELOCITY_PENALTY,
    )

    # Penalty: Distance | Robot <--> Target
    WEIGHT_DISTANCE_ROBOT_TO_TARGET = -16.0
    MAX_DISTANCE_ROBOT_TO_TARGET_PENALTY = -128.0
    penalty_distance_robot_to_target = torch.clamp_min(
        WEIGHT_DISTANCE_ROBOT_TO_TARGET * torch.square(distance_robot_to_target),
        min=MAX_DISTANCE_ROBOT_TO_TARGET_PENALTY,
    )

    # Reward: Distance (linear and angular) | Robot <--> Target (precision rendezvous)
    WEIGHT_PRECISION_RENDEZVOUS = 256.0
    TANH_STD_PRECISION_RENDEZVOUS_POS = 0.025
    TANH_STD_PRECISION_RENDEZVOUS_QUAT = 0.05
    reward_precision_rendezvous = WEIGHT_PRECISION_RENDEZVOUS * (
        1.0
        - torch.tanh(
            (distance_robot_to_target / TANH_STD_PRECISION_RENDEZVOUS_POS)
            + (
                torch.norm(
                    torch.norm(
                        tf_rotmat_robot_to_target
                        - torch.eye(3, device=device).unsqueeze(0),
                        dim=-1,
                    ),
                    dim=-1,
                )
                / TANH_STD_PRECISION_RENDEZVOUS_QUAT
            )
        )
    )

    ##################
    ## Terminations ##
    ##################
    # No termination condition
    termination = torch.zeros(num_envs, dtype=torch.bool, device=device)
    # Truncation
    truncation = (
        episode_length >= max_episode_length
        if truncate_episodes
        else torch.zeros(num_envs, dtype=torch.bool, device=device)
    )

    return StepReturn(
        {
            "state": {
                "tf_rot6d_robot": tf_rot6d_robot,
                "vel_lin_robot": vel_lin_robot,
                "vel_ang_robot": vel_ang_robot,
                "vel_lin_target": vel_lin_target,
                "vel_ang_target": vel_ang_target,
                "tf_pos_robot_to_target": tf_pos_robot_to_target,
                "tf_rot6d_robot_to_target": tf_rot6d_robot_to_target,
            },
            "proprio": {
                "imu_lin_acc": imu_lin_acc,
                "imu_ang_vel": imu_ang_vel,
                "remaining_fuel": remaining_fuel,
            },
        },
        {
            "penalty_action_rate": penalty_action_rate,
            "penalty_fuel_consumption": penalty_fuel_consumption,
            "penalty_angular_velocity": penalty_angular_velocity,
            "penalty_distance_robot_to_target": penalty_distance_robot_to_target,
            "reward_precision_rendezvous": reward_precision_rendezvous,
        },
        termination,
        truncation,
    )



import gc
import signal
from pathlib import Path
from typing import TYPE_CHECKING, Literal, Sequence

import numpy
from isaacsim.simulation_app import SimulationApp
from rl_zoo3 import ALGOS
from stable_baselines3.common.callbacks import tqdm

from srb.integrations.sb3.exp_manager import ExperimentManager
from srb.integrations.sb3.wrapper import Sb3EnvWrapper
from srb.utils import logging
from srb.utils.cfg import last_file, stamp_dir

if TYPE_CHECKING:
    from srb._typing import AnyEnv, AnyEnvCfg

FRAMEWORK_NAME = "sb3"
OFF_POLICY_ALGOS: Sequence[str] = ("qrdqn", "dqn", "ddpg", "sac", "her", "td3", "tqc")


def run(
    workflow: Literal["train", "eval"],
    algo: str,
    env: "AnyEnv",
    sim_app: SimulationApp,
    env_id: str,
    env_cfg: "AnyEnvCfg",
    agent_cfg: dict,
    logdir: Path,
    model: Path,
    continue_training: bool | None = None,
    **kwargs,
):
    ## Extract params from agent_cfg
    # All
    log_interval = agent_cfg.pop("log_interval", -1)
    verbose = agent_cfg.pop("verbose", True)
    track = agent_cfg.pop("track", False)
    init_tensorboard = agent_cfg.pop("tensorboard", True)
    init_wandb = agent_cfg.pop("wandb", False)
    # Train
    save_freq = agent_cfg.pop("save_freq", -1)
    save_replay_buffer = agent_cfg.pop("save_replay_buffer", False)
    # Optimize
    n_trials = agent_cfg.pop("n_trials", 500)
    sampler = agent_cfg.pop("sampler", "tpe")
    pruner = agent_cfg.pop("pruner", "median")
    n_startup_trials = agent_cfg.pop("n_startup_trials", 10)
    n_evaluations = agent_cfg.pop("n_evaluations", 1)
    # HER
    truncate_last_trajectory = agent_cfg.pop("truncate_last_trajectory", True)

    # Determine checkpoint path
    if model:
        from_checkpoint = model
    elif workflow == "eval" or continue_training:
        from_checkpoint = last_file(logdir.joinpath("ckpt"), modification_time=True)
    else:
        from_checkpoint = ""
    if from_checkpoint:
        logging.info(f"Loading model from {from_checkpoint}")

    # Special handling for eval workflow
    if workflow == "eval":
        logdir = stamp_dir(logdir.joinpath("eval"))

    tensorboard_log = (
        logdir.joinpath("tensorboard") if track or init_tensorboard else None
    )
    if init_wandb:
        import wandb

        _run = wandb.init(
            name=f"{env_id}_{algo}",
            sync_tensorboard=True,
            monitor_gym=True,
        )

    # Wrap the environment
    env = Sb3EnvWrapper(env)  # type: ignore

    exp_manager = ExperimentManager(
        args={},  # type: ignore
        algo=algo,
        env_id=env_id,
        log_folder=logdir.as_posix(),
        tensorboard_log=tensorboard_log.as_posix() if tensorboard_log else "",
        n_timesteps=0,
        eval_freq=0,
        n_eval_episodes=0,
        save_freq=save_freq,
        hyperparams=agent_cfg,
        trained_agent=from_checkpoint.as_posix()
        if isinstance(from_checkpoint, Path)
        else from_checkpoint,
        optimize_hyperparameters=workflow == "optimize",
        n_trials=n_trials,
        sampler=sampler,
        pruner=pruner,
        # optimization_log_path=args.optimization_log_path,
        n_startup_trials=n_startup_trials,
        n_evaluations=n_evaluations,
        truncate_last_trajectory=truncate_last_trajectory,
        seed=env_cfg.seed,
        log_interval=log_interval,
        save_replay_buffer=save_replay_buffer,
        verbose=verbose,
        n_eval_envs=0,
        no_optim_plots=False,
        device=env.unwrapped.device,  # type: ignore
        config=None,
        show_progress=True,
        env=env,
    )

    # Run the workflow
    match workflow:
        case "train":
            agent_model, _saved_hyperparams = exp_manager.setup_experiment()  # type: ignore
            exp_manager.learn(agent_model)
            exp_manager.save_trained_model(agent_model)
        case "optimize":
            exp_manager.setup_experiment()
            exp_manager.hyperparameters_optimization()
        case "eval":
            env = exp_manager.create_envs(0, eval_env=True)  # type: ignore

            # Update agent config
            if algo in OFF_POLICY_ALGOS:
                agent_cfg.update(dict(buffer_size=1))
                if "optimize_memory_usage" in agent_cfg:
                    agent_cfg.update(optimize_memory_usage=False)
            if "HerReplayBuffer" in agent_cfg.get("replay_buffer_class", ""):
                agent_cfg["env"] = env

            # Load the agent
            agent = ALGOS[algo].load(
                from_checkpoint.as_posix(),  # type: ignore
                device=env.unwrapped.device,  # type: ignore
            )

            # Initialize the runner
            episode_start = numpy.ones(
                (env.unwrapped.num_envs,),  # type: ignore
                dtype=bool,
            )
            lstm_states = None

            obs = env.reset()
            for _ in tqdm(range(agent_cfg["n_timesteps"])):
                if not sim_app.is_running():
                    break
                action, lstm_states = agent.predict(
                    obs,  # type: ignore
                    state=lstm_states,
                    episode_start=episode_start,  # type: ignore
                    deterministic=True,
                )
                obs, _reward, episode_start, _infos = env.step(action)  # type: ignore


def gc_tqdm(*args):
    tqdm_objects = [obj for obj in gc.get_objects() if "tqdm" in type(obj).__name__]
    for tqdm_object in tqdm_objects:
        if "tqdm_rich" in type(tqdm_object).__name__:
            tqdm_object.close()
    raise KeyboardInterrupt


signal.signal(signal.SIGINT, gc_tqdm)



from __future__ import annotations

from typing import Sequence, Type

from srb.core.action import ActionGroup
from srb.core.asset import ArticulationCfg, RigidObjectCfg
from srb.core.asset.common import Frame
from srb.core.asset.object.object import Object, ObjectRegistry
from srb.core.asset.object.object_type import ObjectType


class Tool(Object, object_entrypoint=ObjectType.TOOL):
    ## Model
    asset_cfg: RigidObjectCfg | ArticulationCfg

    ## Frames
    frame_mount: Frame
    frame_tool_centre_point: Frame

    @classmethod
    def object_registry(cls) -> Sequence[Type[Tool]]:
        return ObjectRegistry.registry.get(ObjectType.TOOL, [])  # type: ignore


class ActiveTool(Tool, object_metaclass=True):
    ## Model
    asset_cfg: ArticulationCfg

    ## Actions
    actions: ActionGroup



from dataclasses import _MISSING_TYPE
from dataclasses import MISSING as MISSING_MANIPULATOR
from dataclasses import MISSING as MISSING_MOBILE_BASE

from srb.core.action import ActionGroup
from srb.core.asset import (
    AerialManipulator,
    AerialRobot,
    ArticulationCfg,
    Frame,
    GroundManipulator,
    GroundRobot,
    Manipulator,
    OrbitalManipulator,
    OrbitalRobot,
    RigidObjectCfg,
)


class GenericAerialManipulator(AerialManipulator, arbitrary_types_allowed=True):
    ## Model
    mobile_base: AerialRobot | None = None
    manipulator: Manipulator | _MISSING_TYPE = MISSING_MANIPULATOR

    ## Other fields are updated via `mobile_base`
    asset_cfg: RigidObjectCfg | ArticulationCfg | _MISSING_TYPE = MISSING_MOBILE_BASE
    actions: ActionGroup | _MISSING_TYPE = MISSING_MOBILE_BASE
    frame_base: Frame | _MISSING_TYPE = MISSING_MOBILE_BASE
    frame_payload_mount: Frame | _MISSING_TYPE = MISSING_MOBILE_BASE
    frame_manipulator_mount: Frame | _MISSING_TYPE = MISSING_MOBILE_BASE
    frame_onboard_camera: Frame | _MISSING_TYPE = MISSING_MOBILE_BASE


class GenericGroundManipulator(GroundManipulator, arbitrary_types_allowed=True):
    ## Model
    mobile_base: GroundRobot | None = None
    manipulator: Manipulator | _MISSING_TYPE = MISSING_MANIPULATOR

    ## Other fields are updated via `mobile_base`
    asset_cfg: RigidObjectCfg | ArticulationCfg | _MISSING_TYPE = MISSING_MOBILE_BASE
    actions: ActionGroup | _MISSING_TYPE = MISSING_MOBILE_BASE
    frame_base: Frame | _MISSING_TYPE = MISSING_MOBILE_BASE
    frame_payload_mount: Frame | _MISSING_TYPE = MISSING_MOBILE_BASE
    frame_manipulator_mount: Frame | _MISSING_TYPE = MISSING_MOBILE_BASE
    frame_front_camera: Frame | _MISSING_TYPE = MISSING_MOBILE_BASE


class GenericOrbitalManipulator(OrbitalManipulator, arbitrary_types_allowed=True):
    ## Model
    mobile_base: OrbitalRobot | None = None
    manipulator: Manipulator | _MISSING_TYPE = MISSING_MANIPULATOR

    ## Other fields are updated via `mobile_base`
    asset_cfg: RigidObjectCfg | ArticulationCfg | _MISSING_TYPE = MISSING_MOBILE_BASE
    actions: ActionGroup | _MISSING_TYPE = MISSING_MOBILE_BASE
    frame_base: Frame | _MISSING_TYPE = MISSING_MOBILE_BASE
    frame_payload_mount: Frame | _MISSING_TYPE = MISSING_MOBILE_BASE
    frame_manipulator_mount: Frame | _MISSING_TYPE = MISSING_MOBILE_BASE
    frame_onboard_camera: Frame | _MISSING_TYPE = MISSING_MOBILE_BASE



from .differential_ik import (  # noqa: F401
    DifferentialInverseKinematicsAction,
    DifferentialInverseKinematicsActionCfg,
)
from .osc import (  # noqa: F401
    OperationalSpaceControllerAction,
    OperationalSpaceControllerActionCfg,
)



from functools import cached_property
from typing import Dict, Sequence, Tuple

import gymnasium
import numpy
import torch
from isaaclab.envs import DirectRLEnv as __DirectRLEnv

from srb._typing import StepReturn
from srb.core.asset import Articulation, RigidObject
from srb.core.manager import ActionManager
from srb.core.sim.robot_setup import AssembledBodies, RobotAssembler
from srb.utils import logging
from srb.utils.math import combine_frame_transforms, subtract_frame_transforms
from srb.utils.str import resolve_env_prim_path

from .cfg import DirectEnvCfg


class __PostInitCaller(type):
    def __call__(cls, *args, **kwargs):
        obj = type.__call__(cls, *args, **kwargs)
        obj.__post_init__()
        return obj


class DirectEnv(__DirectRLEnv, metaclass=__PostInitCaller):
    cfg: DirectEnvCfg
    _step_return: StepReturn

    def __init__(self, cfg: DirectEnvCfg, **kwargs):
        super().__init__(cfg, **kwargs)

        # Apply visuals
        self.cfg.visuals.func(self.cfg.visuals)

        # Add action manager
        if self.cfg.actions:
            self.action_manager = ActionManager(
                self.cfg.actions,
                env=self,  # type: ignore
            )
            print(self.action_manager)

        ## Get scene assets
        self._robot: Articulation = self.scene["robot"]

    def close(self):
        if not self._is_closed:
            if self.cfg.actions:
                del self.action_manager

        super().close()

    def extract_step_return(self) -> StepReturn:
        raise NotImplementedError

    def _extract_step_return_wrapped(self) -> StepReturn:
        step_return = self.extract_step_return()

        ## Detect environments that "exploded" due to physics and terminate them
        invalid_envs = torch.zeros(self.num_envs, dtype=torch.bool, device=self.device)
        invalid_keys = []
        for rew_key, rew_val in step_return.reward.items():
            rew_invalid = ~torch.isfinite(rew_val)
            invalid_envs = invalid_envs | rew_invalid
            if rew_invalid.any():
                step_return.reward[rew_key] = torch.where(
                    rew_invalid, torch.zeros_like(rew_val), rew_val
                )
                invalid_keys.append(rew_key)
        for obs_cat, obs_group in step_return.observation.items():
            for obs_key, obs_val in obs_group.items():
                if len(obs_val.shape) > 1:
                    dims = tuple(range(1, len(obs_val.shape)))
                    obs_invalid = torch.any(~torch.isfinite(obs_val), dim=dims)
                else:
                    obs_invalid = ~torch.isfinite(obs_val)
                invalid_envs = invalid_envs | obs_invalid
                if obs_invalid.any():
                    mask = obs_invalid
                    for _ in range(len(obs_val.shape) - 1):
                        mask = mask.unsqueeze(-1)
                    mask = mask.expand_as(obs_val)
                    step_return.observation[obs_cat][obs_key] = torch.where(
                        mask, torch.zeros_like(obs_val), obs_val
                    )
                    invalid_keys.append(f"{obs_cat}/{obs_key}")
        if invalid_envs.any():
            logging.warning(
                f"Terminating {invalid_envs.sum().item()} environment instance{'s' if invalid_envs.sum().item() > 1 else ''} due to non-finite rewards or observations: {', '.join(invalid_keys)}"
            )
            step_return = StepReturn(
                observation=step_return.observation,
                reward=step_return.reward,
                termination=torch.where(
                    invalid_envs, torch.ones_like(invalid_envs), step_return.termination
                ),
                truncation=step_return.truncation,
                info=step_return.info,
            )

        return step_return

    def _reset_idx(self, env_ids: Sequence[int]):
        if self.cfg.actions:
            self.action_manager.reset(env_ids)

        super()._reset_idx(env_ids)

        # Move assembled bodies to the correct position to avoid physics snapping them in place
        self._update_assembly_fixed_joint_transforms(env_ids)

    def _pre_physics_step(self, actions: torch.Tensor):
        if self.cfg.actions:
            self.action_manager.process_action(actions)
        else:
            super()._pre_physics_step(actions)  # type: ignore

    def _apply_action(self):
        if self.cfg.actions:
            self.action_manager.apply_action()
        else:
            super()._apply_action()  # type: ignore

    def __post_init__(self):
        if self._use_step_return_workflow:
            self._step_return = self.extract_step_return()

            # Verify that all observation components have the correct shape and are finite
            for obs_cat, obs_group in self._step_return.observation.items():
                for obs_key, obs_val in obs_group.items():
                    assert obs_val.size(0) == self.num_envs, (
                        f"Observation component '{obs_cat}/{obs_key}' has an incorrect shape. "
                        f"Expected: ({self.num_envs}, ...) | Actual: {obs_val.shape}"
                    )
                    assert torch.isfinite(obs_val).all(), (
                        f"Observation component '{obs_cat}/{obs_key}' contains non-finite values."
                    )
            # Verify that all reward components have the correct shape and are finite
            for rew_key, rew_val in self._step_return.reward.items():
                assert rew_val.shape == (self.num_envs,), (
                    f"Reward component '{rew_key}' has an incorrect shape. "
                    f"Expected: ({self.num_envs},) | Actual: {rew_val.shape}"
                )
                assert torch.isfinite(rew_val).all(), (
                    f"Reward component '{rew_key}' contains non-finite values."
                )

        # Automatically determine the action and observation spaces for all sub-classes
        self._update_gym_env_spaces()

    def _update_gym_env_spaces(self):
        # Action space
        self.single_action_space = gymnasium.spaces.Box(
            low=-1.0,
            high=1.0,
            shape=(self.action_manager.total_action_dim,),
        )
        self.action_space = gymnasium.vector.utils.batch_space(
            self.single_action_space, self.num_envs
        )

        # Observation space
        self.single_observation_space = gymnasium.spaces.Dict({})
        for obs_key, obs_buf in self._get_observations().items():
            assert isinstance(obs_buf, (numpy.ndarray, torch.Tensor))
            if isinstance(obs_buf, torch.Tensor):
                match obs_buf.dtype:
                    case torch.float32:
                        dtype = numpy.float32
                    case torch.float64:
                        dtype = numpy.float64
                    case torch.int8:
                        dtype = numpy.int8
                    case torch.uint8:
                        dtype = numpy.uint8
                    case torch.int16:
                        dtype = numpy.int16
                    case torch.uint16:
                        dtype = numpy.uint16
                    case torch.int32:
                        dtype = numpy.int32
                    case torch.uint32:
                        dtype = numpy.uint32
                    case torch.int64:
                        dtype = numpy.int64
                    case torch.uint64:
                        dtype = numpy.uint64
                    case torch.bool:
                        dtype = numpy.bool_
                    case _:
                        raise TypeError(
                            f"Unsupported torch dtype '{obs_buf.dtype}' for observation '{obs_key}'."
                        )
            else:
                dtype = obs_buf.dtype
            match dtype:
                case numpy.float32 | numpy.float64:
                    low = numpy.finfo(dtype).min  # type: ignore
                    high = numpy.finfo(dtype).max  # type: ignore
                case (
                    numpy.int8
                    | numpy.uint8
                    | numpy.int16
                    | numpy.uint16
                    | numpy.int32
                    | numpy.uint32
                    | numpy.int64
                    | numpy.uint64
                ):
                    low = numpy.iinfo(dtype).min
                    high = numpy.iinfo(dtype).max
                case numpy.bool_:
                    low = 0
                    high = 1
                case _:
                    low = -numpy.inf
                    high = numpy.inf
            self.single_observation_space[obs_key] = gymnasium.spaces.Box(
                low=low,
                high=high,
                shape=obs_buf.shape[1:],
                dtype=dtype,  # type: ignore
            )
        self.observation_space = gymnasium.vector.utils.batch_space(
            self.single_observation_space, self.num_envs
        )

    @cached_property
    def max_episode_length(self):
        # Wrap lengthy calculation in a cached property
        return super().max_episode_length

    @cached_property
    def _use_step_return_workflow(self) -> bool:
        return self.__class__.extract_step_return is not DirectEnv.extract_step_return

    def _get_dones(self) -> Tuple[torch.Tensor, torch.Tensor]:
        if self._use_step_return_workflow:
            self._step_return = self._extract_step_return_wrapped()
            if self.cfg.extras:
                self.extras["reward_terms"] = self._step_return.reward
            if self._step_return.info:
                self.extras.update(self._step_return.info)
            return self._step_return.termination, self._step_return.truncation
        else:
            return super()._get_dones()  # type: ignore

    def _get_rewards(self) -> torch.Tensor:
        if self._use_step_return_workflow:
            return _sum_rewards(self._step_return.reward)
        else:
            return super()._get_rewards()  # type: ignore

    def _get_observations(self) -> Dict[str, torch.Tensor]:
        if self._use_step_return_workflow:
            return _flatten_observations(self._step_return.observation)
        else:
            return super()._get_observations()  # type: ignore

    def _setup_scene(self):
        super()._setup_scene()

        ## Handle assemblies
        self.joint_assemblies: Dict[str, Sequence[AssembledBodies]] = {}
        for key, assembly_cfg in self.cfg.joint_assemblies.items():
            self.joint_assemblies[key] = tuple(
                RobotAssembler().assemble_rigid_bodies(
                    assembly_cfg.model_copy(
                        update={
                            "base_path": resolve_env_prim_path(
                                assembly_cfg.base_path, i
                            ),
                            "attach_path": resolve_env_prim_path(
                                assembly_cfg.attach_path, i
                            ),
                        }
                    )
                )
                for i in range(self.num_envs)
            )

    def _update_assembly_fixed_joint_transforms(self, env_ids: Sequence[int]):
        for key, assembly_cfg in self.cfg.joint_assemblies.items():
            attach_asset: RigidObject | Articulation = self.scene[key]
            base_asset: RigidObject | Articulation = self.scene[
                assembly_cfg.base_path.rsplit("/", 1)[-1]
            ]
            attach_root_state = attach_asset.data.root_state_w[env_ids]
            attach_body_state = (
                attach_asset.data.body_state_w[
                    env_ids,
                    attach_asset.find_bodies(
                        assembly_cfg.attach_mount_frame.removeprefix("/")
                    )[0][0],
                ]
                if assembly_cfg.attach_mount_frame
                else attach_root_state
            )
            base_body_state = (
                base_asset.data.body_state_w[
                    env_ids,
                    base_asset.find_bodies(
                        assembly_cfg.base_mount_frame.removeprefix("/")
                    )[0][0],
                ]
                if assembly_cfg.base_mount_frame
                else base_asset.data.root_state_w[env_ids]
            )

            pose = torch.cat(
                combine_frame_transforms(
                    base_body_state[:, :3],
                    base_body_state[:, 3:7],
                    *combine_frame_transforms(
                        *subtract_frame_transforms(
                            attach_body_state[:, :3],
                            attach_body_state[:, 3:7],
                            attach_root_state[:, :3],
                            attach_root_state[:, 3:7],
                        ),
                        torch.tensor(
                            assembly_cfg.fixed_joint_offset,
                            dtype=attach_root_state.dtype,
                            device=attach_root_state.device,
                        )
                        .unsqueeze(0)
                        .repeat(attach_root_state.shape[0], 1),
                        torch.tensor(
                            assembly_cfg.fixed_joint_orient,
                            dtype=attach_root_state.dtype,
                            device=attach_root_state.device,
                        )
                        .unsqueeze(0)
                        .repeat(attach_root_state.shape[0], 1),
                    ),
                ),
                dim=1,
            )

            attach_asset.write_root_pose_to_sim(pose, env_ids=env_ids)


@torch.jit.script
def _flatten_observations(
    obs_dict: Dict[str, Dict[str, torch.Tensor]],
) -> Dict[str, torch.Tensor]:
    return {
        obs_cat: torch.cat(
            [torch.flatten(tensor, start_dim=1) for tensor in obs_group.values()], dim=1
        )
        for obs_cat, obs_group in obs_dict.items()
    }


@torch.jit.script
def _sum_rewards(rew_dict: Dict[str, torch.Tensor]) -> torch.Tensor:
    return torch.sum(torch.stack(list(rew_dict.values()), dim=1), dim=1)



import os
import subprocess
import tempfile
import time
from pathlib import Path
from typing import Iterable

import pytest

from srb.utils import logging
from srb.utils.cache import read_offline_srb_env_cache
from srb.utils.isaacsim import get_isaacsim_python
from srb.utils.subprocess import terminate_process

NUM_ENVS: int = 4
MAX_DURATION_RAND: float = 40.0
MAX_DURATION_TRAIN: float = 80.0

HEADLESS: bool = True
STACK: bool = True
TEST_VISUAL_ENVS: bool = False
TEST_TEMPLATE_ENVS: bool = False
TEST_EXCAVATION: bool = True

PERF: bool = False
PERF_DURATION: float = 600.0
PERF_OUTPUT_DIR: str = os.path.join(
    os.path.dirname(os.path.dirname(os.path.abspath(__file__))), "logs", "_perf"
)

TEST_START_TIME: float = time.time()


def _list_envs() -> Iterable[str]:
    if envs := read_offline_srb_env_cache():
        if not TEST_VISUAL_ENVS:
            envs = filter(lambda env: not _is_visual_env(env), envs)
        if not TEST_TEMPLATE_ENVS:
            envs = filter(lambda env: not env.startswith("_"), envs)
        if not TEST_EXCAVATION:
            envs = filter(lambda env: "excavation" not in env, envs)
        envs = sorted(envs)
    else:
        test_filepath = Path(__file__)
        logging.warning(
            f"Skipping {test_filepath.parent.name}/{test_filepath.name} because no environments were found in the cache (repeat the test)"
        )
        envs = []

    return envs


def _is_visual_env(env: str) -> bool:
    return env.endswith("_visual")


@pytest.mark.parametrize(
    "env,num_envs",
    ((env, num_envs) for env in _list_envs() for num_envs in (NUM_ENVS,)),
)
def test_cli_agent_rand(env: str, num_envs: int):
    with tempfile.TemporaryDirectory(prefix="srb_") as tmpdir:
        cmd = (
            get_isaacsim_python(),
            "-m",
            "srb",
            "agent",
            "rand",
            "--headless" if HEADLESS else "--hide_ui",
            f"env.scene.num_envs={num_envs}",
            "--env",
            env,
            "--logs",
            tmpdir,
            f"env.stack={STACK}",
        )

        if PERF:
            cmd += (
                "--perf",
                "--perf_duration",
                str(PERF_DURATION),
                "--perf_output",
                os.path.join(
                    PERF_OUTPUT_DIR,
                    f"perf_{env.replace('/', '_')}_{num_envs}_{int(TEST_START_TIME)}.txt",
                ),
            )

            if not os.path.exists(PERF_OUTPUT_DIR):
                os.makedirs(PERF_OUTPUT_DIR, exist_ok=True)
            with open(
                os.path.join(PERF_OUTPUT_DIR, f"commands_{int(TEST_START_TIME)}.bash"),
                "a",
            ) as f:
                f.write(" ".join(cmd) + "\n")

        environ = os.environ.copy()
        environ["SF_BAKER"] = "0"

        process = None
        try:
            process = subprocess.Popen(
                cmd,
                env=environ,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                text=True,
                preexec_fn=os.setsid,
            )

            if PERF:
                process.wait()
            else:
                start = time.time()
                while time.time() - start < MAX_DURATION_RAND:
                    if process.poll() is not None:
                        logging.critical(f"[{env}] Failed command: {' '.join(cmd)}")
                        stdout, stderr = process.communicate()
                        pytest.fail(
                            f'Process failed for env "{env}"\n[env={env}] STDOUT:\n{stdout}\n[env={env}] STDERR:\n{stderr}'
                        )
                    time.sleep(0.01 * MAX_DURATION_RAND)
        except Exception as e:
            logging.critical(f"[{env}] Failed command: {' '.join(cmd)}")
            pytest.fail(
                f'Failed to start process for env "{env}"\n[env={env}] Exception: {e}'
            )
        finally:
            terminate_process("srb", process)


@pytest.mark.skip
@pytest.mark.parametrize(
    "env,num_envs,algo",
    (
        (env, num_envs, algo)
        for env in _list_envs()
        for num_envs in (NUM_ENVS,)
        for algo in ("dreamer",)
    ),
)
def test_cli_agent_train(env: str, num_envs: int, algo: str):
    with tempfile.TemporaryDirectory(prefix="srb_") as tmpdir:
        match algo:
            case _:
                agent_params = ()

        cmd = (
            get_isaacsim_python(),
            "-m",
            "srb",
            "agent",
            "train",
            "--headless" if HEADLESS else "--hide_ui",
            f"env.scene.num_envs={num_envs}",
            "--env",
            env,
            "--algo",
            algo,
            "--logs",
            tmpdir,
            *agent_params,
        )

        environ = os.environ.copy()
        environ["SF_BAKER"] = "0"

        process = None
        try:
            process = subprocess.Popen(
                cmd,
                env=environ,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                text=True,
                preexec_fn=os.setsid,
            )

            start = time.time()
            while time.time() - start < MAX_DURATION_TRAIN:
                if process.poll() is not None:
                    logging.critical(f"[{env}] Failed command: {' '.join(cmd)}")
                    stdout, stderr = process.communicate()
                    pytest.fail(
                        f'Process failed for env "{env}"\n[env={env}] STDOUT:\n{stdout}\n[env={env}] STDERR:\n{stderr}'
                    )
                time.sleep(0.01 * MAX_DURATION_TRAIN)
        except Exception as e:
            logging.critical(f"[{env}] Failed command: {' '.join(cmd)}")
            pytest.fail(
                f'Failed to start process for env "{env}"\n[env={env}] Exception: {e}'
            )
        finally:
            terminate_process("srb", process)


# TODO[mid]: Add test for combination of all robots and end-effectors



from srb.core.env.common.extension.visual import VisualExtCfg
from srb.core.sensor import CameraCfg, PinholeCameraCfg
from srb.utils.cfg import configclass
from srb.utils.math import rpy_to_quat  # noqa: F401

from .env import AerialEnvCfg


@configclass
class AerialEnvVisualExtCfg(VisualExtCfg):
    def wrap(self, env_cfg: AerialEnvCfg):
        self.cameras_cfg = {
            # "cam_scene": CameraCfg(
            #     prim_path=f"{env_cfg._robot.asset_cfg.prim_path}{('/' + env_cfg._robot.frame_base.prim_relpath) if env_cfg._robot.frame_base.prim_relpath else ''}/camera_scene",
            #     offset=CameraCfg.OffsetCfg(
            #         convention="world",
            #         pos=(-2.5, 0.0, 2.5),
            #         rot=rpy_to_quat(0.0, 45.0, 0.0),
            #     ),
            #     spawn=PinholeCameraCfg(
            #         clipping_range=(0.1, 75.0 + 0.1),
            #     ),
            # ),
            "cam_onboard": CameraCfg(
                prim_path=f"{env_cfg._robot.asset_cfg.prim_path}/{env_cfg._robot.frame_onboard_camera.prim_relpath}",
                offset=CameraCfg.OffsetCfg(
                    convention="world",
                    pos=env_cfg._robot.frame_onboard_camera.offset.pos,
                    rot=env_cfg._robot.frame_onboard_camera.offset.rot,
                ),
                spawn=PinholeCameraCfg(
                    focal_length=5.0,
                    horizontal_aperture=12.0,
                    clipping_range=(0.05, 50.0 + 0.05),
                ),
            ),
        }

        super().wrap(env_cfg)



from .extravehicular import ExtravehicularScenery  # noqa: F401
from .intravehicular import IntravehicularScenery  # noqa: F401
from .scenery import Scenery, SceneryRegistry  # noqa: F401
from .scenery_type import SceneryType  # noqa: F401
from .subterrane import Subterrane  # noqa: F401
from .terrain import Terrain  # noqa: F401



from typing import TYPE_CHECKING

import gymnasium
import numpy
import rl_zoo3
import rl_zoo3.enjoy
import rl_zoo3.train
import torch
from gymnasium.core import ObservationWrapper, ObsType, WrapperObsType
from rl_zoo3 import exp_manager
from sbx import DDPG, DQN, PPO, SAC, TD3, TQC, CrossQ

from srb.integrations.sb3 import main as sb3

if TYPE_CHECKING:
    from srb._typing import AnyEnv


sb3.FRAMEWORK_NAME = "sbx"

rl_zoo3.ALGOS["ddpg"] = DDPG
rl_zoo3.ALGOS["dqn"] = DQN
rl_zoo3.ALGOS["sac"] = SAC
rl_zoo3.ALGOS["ppo"] = PPO
rl_zoo3.ALGOS["td3"] = TD3
rl_zoo3.ALGOS["tqc"] = TQC
rl_zoo3.ALGOS["crossq"] = CrossQ
rl_zoo3.train.ALGOS = rl_zoo3.ALGOS
rl_zoo3.enjoy.ALGOS = rl_zoo3.ALGOS
exp_manager.ALGOS = rl_zoo3.ALGOS


def run(env: "AnyEnv", **kwargs):
    # SBX does not support MultiInputPolicy, so merge all obs from dict into a single box
    env = MergeDictObsWrapper(env)  # type: ignore

    sb3.run(env=env, **kwargs)


class MergeDictObsWrapper(ObservationWrapper):
    def __init__(self, env):
        super().__init__(env)

        assert isinstance(
            env.unwrapped.single_observation_space,  # type: ignore
            gymnasium.spaces.Dict,
        ), "The observation space must be of type gymnasium.spaces.Dict"

        self.dict_keys = tuple(env.unwrapped.single_observation_space.spaces.keys())  # type: ignore

        # Collect all the low and high bounds for the new Box
        low = []
        high = []
        for key in self.dict_keys:
            space = env.unwrapped.single_observation_space.spaces[key]  # type: ignore
            space_length = numpy.prod(space.shape)  # type: ignore
            space_low = space.low  # type: ignore
            space_high = space.high  # type: ignore

            if isinstance(space_low, numpy.ndarray) and space_length > 1:
                low.extend(space_low.flatten())
                high.extend(space_high.flatten())
            else:
                low.extend((space_low,) * space_length)
                high.extend((space_high,) * space_length)

        # Define the new merged observation space
        self.unwrapped.single_observation_space = gymnasium.spaces.Box(  # type: ignore
            low=numpy.array(low),
            high=numpy.array(high),
            dtype=numpy.float32,
        )

        # Ensure the correct observation and action spaces are set
        self._observation_space = self.unwrapped.single_observation_space  # type: ignore
        self._action_space = self.unwrapped.single_action_space  # type: ignore

    def observation(self, observation: ObsType) -> WrapperObsType:  # type: ignore
        return torch.cat(  # type: ignore
            [
                torch.flatten(
                    observation[key],  # type: ignore
                    start_dim=1,
                )
                for key in self.dict_keys
            ],
            dim=1,
        )



from srb.utils import logging
from srb.utils.importer import import_recursively
from srb.utils.isaacsim import is_isaacsim_initialized
from srb.utils.registry import get_srb_tasks

if is_isaacsim_initialized():
    import_recursively(__name__)
    logging.debug(
        f'Recursively imported Space Robotics Bench module "{__name__}" ({len(get_srb_tasks())} registered tasks)'
    )
else:
    raise RuntimeError(
        "Tasks of the Space Robotics Bench cannot be registered because Isaac Sim is not initialized. "
        f'Please import the "{__name__}" module after starting the Omniverse simulation app.'
    )



from .light import Light  # noqa: F401
from .object import Object, ObjectRegistry  # noqa: F401
from .object_type import ObjectType  # noqa: F401
from .payload import Payload  # noqa: F401
from .pedestal import Pedestal  # noqa: F401
from .tool import ActiveTool, Tool  # noqa: F401



from .combined import *  # noqa: F403
from .humanoid import Humanoid  # noqa: F401
from .mobile_manipulator import (  # noqa: F401
    MobileManipulator,
    MobileManipulatorRegistry,
)
from .mobile_manipulator_type import MobileManipulatorType  # noqa: F401



from isaaclab.envs import ManagerBasedRLEnv as __ManagerBasedRLEnv

from .cfg import ManagedEnvCfg


class ManagedEnv(__ManagerBasedRLEnv):
    cfg: ManagedEnvCfg

    def __init__(self, cfg: ManagedEnvCfg, **kwargs):
        super().__init__(cfg, **kwargs)

        # Apply visuals
        self.cfg.visuals.func(self.cfg.visuals)



import gymnasium
from isaaclab.envs import DirectRLEnvCfg as __DirectRLEnvCfg

from srb.utils.cfg import configclass

from ..env_cfg import BaseEnvCfg


@configclass
class DirectEnvCfg(BaseEnvCfg, __DirectRLEnvCfg):
    # Disable UI window by default
    ui_window_class_type: type | None = None

    # Temporarily set action_space, observation_space, and state_space (overridden by the implementation)
    # TODO[low]: Clean-up DirectEnv patch
    action_space = gymnasium.spaces.Box(low=-1.0, high=1.0, shape=(1,))
    observation_space = gymnasium.spaces.Box(low=-1.0, high=1.0, shape=(1,))
    state_space = gymnasium.spaces.Box(low=-1.0, high=1.0, shape=(1,))

    def __post_init__(self):
        BaseEnvCfg.__post_init__(self)



from typing import TYPE_CHECKING, Sequence, Tuple, Type

import torch
from pydantic import BaseModel

from srb.core.manager import ActionTerm, ActionTermCfg
from srb.core.marker import ARROW_CFG, VisualizationMarkers
from srb.core.sim import PreviewSurfaceCfg
from srb.utils.cfg import configclass
from srb.utils.math import (
    combine_frame_transforms,
    matrix_from_euler,
    normalize,
    quat_apply,
    quat_from_matrix,
)

if TYPE_CHECKING:
    from srb._typing import AnyEnv
    from srb.core.asset import RigidObject


class ThrustAction(ActionTerm):
    cfg: "ThrustActionCfg"
    _env: "AnyEnv"
    _asset: "RigidObject"

    def __init__(self, cfg: "ThrustActionCfg", env: "AnyEnv"):
        super().__init__(
            cfg,
            env,  # type: ignore
        )

        ## Pre-process thrusters
        thruster_offset = []
        thruster_direction = []
        thruster_power = []
        thruster_gimbal_limits = []
        for thruster_cfg in cfg.thrusters:
            thruster_offset.append(thruster_cfg.offset)
            direction_norm = (
                thruster_cfg.direction[0] ** 2
                + thruster_cfg.direction[1] ** 2
                + thruster_cfg.direction[2] ** 2
            ) ** 0.5
            assert direction_norm > 0, (
                "Thruster direction must have a non-zero magnitude"
            )
            direction = (
                thruster_cfg.direction[0] / direction_norm,
                thruster_cfg.direction[1] / direction_norm,
                thruster_cfg.direction[2] / direction_norm,
            )
            thruster_direction.append(direction)
            thruster_gimbal_limits.append(thruster_cfg.gimbal_limits or (0.0, 0.0))
            thruster_power.append(thruster_cfg.power)
        self._thruster_offset = torch.tensor(thruster_offset, device=env.device)
        self._thruster_direction = torch.tensor(thruster_direction, device=env.device)
        self._thruster_power = torch.tensor(thruster_power, device=env.device)
        self._thruster_gimbal_limits = torch.tensor(
            thruster_gimbal_limits, device=env.device
        )

        ## Set up action indices
        self._num_thrusters = len(cfg.thrusters)
        self._action_indices_thrust = torch.arange(
            self._num_thrusters, device=env.device
        )
        self._num_thrusters_with_gimbal = len(
            [limits for limits in thruster_gimbal_limits if limits != (0.0, 0.0)]
        )
        action_indices_gimbal: Sequence[Tuple[int, int]] = []
        for i, limits in enumerate(thruster_gimbal_limits):
            if limits == (0.0, 0.0):
                action_indices_gimbal.append((-1, -1))
            else:
                start_idx = self._num_thrusters + 2 * i
                action_indices_gimbal.append((start_idx, start_idx + 1))
        self._action_indices_gimbal = torch.tensor(
            action_indices_gimbal, device=env.device
        )

        ## Initialize fuel & mass
        self._remaining_fuel = cfg.fuel_capacity * torch.ones(
            env.num_envs, device=env.device
        )
        self._dry_masses = self._asset.root_physx_view.get_masses().clone()

        ## Set up visualization markers
        if self.cfg.debug_vis:
            self._setup_visualization_markers()

    @property
    def action_dim(self) -> int:
        return self._num_thrusters + 2 * self._num_thrusters_with_gimbal

    @property
    def raw_actions(self) -> torch.Tensor:
        return self._raw_actions

    @property
    def processed_actions(self) -> torch.Tensor:
        return self._processed_actions

    @property
    def remaining_fuel(self) -> torch.Tensor:
        return self._remaining_fuel

    def process_actions(self, actions):
        self._raw_actions = actions
        self._processed_actions = actions.clone()
        self._processed_actions[:, self._action_indices_thrust] = torch.clamp(
            self._processed_actions[:, self._action_indices_thrust], 0.0, 1.0
        )

    def apply_actions(self):
        ## Apply gimbal rotations to thruster directions
        thruster_directions = (
            self._thruster_direction.unsqueeze(0).expand(self.num_envs, -1, -1).clone()
        )
        for i in range(self._num_thrusters):
            gimbal_i = self._action_indices_gimbal[i]

            # Skip thrusters without gimbal
            if gimbal_i[0] < 0:
                continue

            # Extract gimbal actions and map them to limits
            lim = self._thruster_gimbal_limits[i]
            gimbal_x = lim[0] * self._processed_actions[:, gimbal_i[0]].clamp(-1.0, 1.0)
            gimbal_y = lim[1] * self._processed_actions[:, gimbal_i[1]].clamp(-1.0, 1.0)

            # Convert gimbal angles to rotation matrices
            euler_angles = torch.zeros((self.num_envs, 3), device=self.device)
            euler_angles[:, 0] = gimbal_x
            euler_angles[:, 1] = gimbal_y
            rotation_matrices = matrix_from_euler(euler_angles, convention="XYZ")

            # Rotate the direction vectors
            nominal_dir = (
                self._thruster_direction[i].unsqueeze(0).expand(self.num_envs, -1)
            )
            rotated_dir = torch.bmm(
                rotation_matrices, nominal_dir.unsqueeze(2)
            ).squeeze(2)

            # Update the direction for this thruster
            thruster_directions[:, i, :] = rotated_dir

        ## Compute thrust force magnitude for each thruster
        thrust_actions = self._processed_actions[:, self._action_indices_thrust]
        thrust_magnitudes = thrust_actions * self._thruster_power.unsqueeze(0)

        ## Disable thrust if fuel is depleted
        thrust_magnitudes *= self._remaining_fuel.unsqueeze(-1) > 0.0

        ## Compute force vector for each thruster
        # Note: The force is applied in the opposite direction of the thrust vector
        thruster_forces = (
            -self.cfg.scale * thrust_magnitudes.unsqueeze(-1) * thruster_directions
            # / self._env.cfg.agent_rate
        )

        ## Get center of mass positions [num_envs, 3]
        thruster_offsets = self._thruster_offset.unsqueeze(0).expand(
            self.num_envs, -1, -1
        )
        com_positions = self._asset.root_physx_view.get_coms()[:, :3].unsqueeze(1)
        thruster_offsets_com = thruster_offsets - com_positions

        ## Calculate torques resulting from thruster forces
        thruster_torques = torch.cross(thruster_offsets_com, thruster_forces, dim=2)

        ## Apply forces and torques at center of mass in the local frame
        self._asset.root_physx_view.apply_forces_and_torques_at_position(
            force_data=thruster_forces.sum(dim=1),
            torque_data=thruster_torques.sum(dim=1),
            position_data=com_positions,
            indices=self._asset._ALL_INDICES,
            is_global=False,
        )

        ## Update fuel and mass
        self._remaining_fuel -= (
            self.cfg.fuel_consumption_rate
            * thrust_magnitudes.sum(dim=1)
            * self._env.cfg.agent_rate
        )
        self._remaining_fuel.clamp_(min=0.0)
        masses = self._dry_masses + self._remaining_fuel.unsqueeze(-1)
        mass_decrease_ratio = masses / self._asset.root_physx_view.get_masses()
        self._asset.root_physx_view.set_masses(masses, indices=self._asset._ALL_INDICES)
        self._asset.root_physx_view.set_inertias(
            mass_decrease_ratio * self._asset.root_physx_view.get_inertias(),
            indices=self._asset._ALL_INDICES,
        )

        ## Update visualization markers
        if self.cfg.debug_vis:
            self._update_visualization_markers(
                thruster_offsets=thruster_offsets,
                thruster_directions=thruster_directions,
                thrust_magnitudes=thrust_magnitudes,
            )

    def _setup_visualization_markers(self):
        # TODO[low]: Support custom visualization markers for thrusters
        self._thruster_markers = []
        for i in range(self._num_thrusters):
            cfg = ARROW_CFG.copy().replace(  # type: ignore
                prim_path=f"/Visuals/thrusters/thruster{i}"
            )
            cfg.markers["arrow"].tail_radius = 0.1
            cfg.markers["arrow"].tail_length = 1.0
            cfg.markers["arrow"].head_radius = 0.2
            cfg.markers["arrow"].head_length = 0.5

            # Use a different color for each thruster (gradient from red to blue)
            blue = i / max(self._num_thrusters - 1, 1)
            cfg.markers["arrow"].visual_material = PreviewSurfaceCfg(
                emissive_color=(1.0 - blue, 0.2, blue)
            )

            # Create the marker and add to list
            self._thruster_markers.append(VisualizationMarkers(cfg))

    def _update_visualization_markers(
        self,
        thruster_offsets: torch.Tensor,
        thruster_directions: torch.Tensor,
        thrust_magnitudes: torch.Tensor,
    ):
        asset_pos = self._asset.data.root_pos_w
        asset_quat = self._asset.data.root_quat_w

        for i in range(self._num_thrusters):
            # Transform thruster position to world frame
            thruster_pos_w, _ = combine_frame_transforms(
                t01=asset_pos,
                q01=asset_quat,
                t12=thruster_offsets[:, i, :],
            )

            # Orient the marker with the thrust vector
            thrust_dir_world = quat_apply(asset_quat, thruster_directions[:, i, :])
            thrust_dir_world = normalize(thrust_dir_world)

            # Create rotation matrix where x-axis is aligned with thrust direction
            x_axis = thrust_dir_world

            # Choose any perpendicular vector for y-axis
            y_axis = torch.zeros_like(x_axis)
            # Find index of smallest component in x_axis to create orthogonal vector
            min_idx = torch.argmin(torch.abs(x_axis), dim=1)
            for env_idx in range(self.num_envs):
                y_axis[env_idx, min_idx[env_idx]] = 1.0

            # Make y_axis perpendicular to x_axis
            y_axis = normalize(
                y_axis - x_axis * torch.sum(x_axis * y_axis, dim=1, keepdim=True)
            )

            # Get z_axis from cross product
            z_axis = normalize(torch.cross(x_axis, y_axis, dim=1))

            # Create rotation matrix
            rot_matrix = torch.zeros((self.num_envs, 3, 3), device=self.device)
            rot_matrix[:, :, 0] = x_axis
            rot_matrix[:, :, 1] = y_axis
            rot_matrix[:, :, 2] = z_axis

            # Convert to quaternion
            thrust_dir_quat_w = quat_from_matrix(rot_matrix)

            # Scale the marker based on the thrust magnitude
            marker_scale = torch.ones((self.num_envs, 3), device=self.device)
            marker_scale[:, :] = (
                thrust_magnitudes[:, i] * (1.0 / self._thruster_power.max().item())
            ).unsqueeze(1)

            # Visualize the marker
            self._thruster_markers[i].visualize(
                thruster_pos_w, thrust_dir_quat_w, marker_scale
            )

    def reset(self, env_ids: Sequence[int] | None = None):
        super().reset(env_ids)
        self._remaining_fuel[env_ids] = self.cfg.fuel_capacity


class ThrusterCfg(BaseModel):
    offset: Tuple[float, float, float] = (0.0, 0.0, 0.0)
    direction: Tuple[float, float, float] = (0.0, 0.0, -1.0)
    gimbal_limits: Tuple[float, float] | None = None
    power: float = 1.0


@configclass
class ThrustActionCfg(ActionTermCfg):
    class_type: Type = ThrustAction
    scale: float = 1.0

    thrusters: Sequence[ThrusterCfg] = (ThrusterCfg(),)
    fuel_capacity: float = 1.0
    fuel_consumption_rate: float = 0.1



from dataclasses import MISSING
from typing import Sequence

import torch

from srb import assets
from srb._typing import StepReturn
from srb.core.asset import (
    Articulation,
    ArticulationCfg,
    AssetBaseCfg,
    AssetVariant,
    Manipulator,
    Object,
    RigidObject,
    RigidObjectCfg,
)
from srb.core.domain import Domain
from srb.core.env import (
    ManipulationEnv,
    ManipulationEnvCfg,
    ManipulationEventCfg,
    ManipulationSceneCfg,
)
from srb.core.manager import EventTermCfg, SceneEntityCfg
from srb.core.mdp import reset_root_state_uniform
from srb.core.sensor import ContactSensor, ContactSensorCfg
from srb.core.sim import (
    ArticulationRootPropertiesCfg,
    CollisionPropertiesCfg,
    MeshCollisionPropertiesCfg,
    RigidBodyPropertiesCfg,
    UsdFileCfg,
)
from srb.utils.cfg import configclass
from srb.utils.math import (
    combine_frame_transforms,
    matrix_from_quat,
    rotmat_to_rot6d,
    rpy_to_quat,
    scale_transform,
    subtract_frame_transforms,
)

##############
### Config ###
##############


@configclass
class SceneCfg(ManipulationSceneCfg):
    bolt: RigidObjectCfg = MISSING  # type: ignore
    nut: RigidObjectCfg = MISSING  # type: ignore

    decor: AssetBaseCfg = AssetBaseCfg(
        prim_path="{ENV_REGEX_NS}/decor",
        spawn=UsdFileCfg(
            usd_path=assets.Ingenuity().asset_cfg.spawn.usd_path,  # type: ignore
            collision_props=CollisionPropertiesCfg(collision_enabled=False),
            mesh_collision_props=MeshCollisionPropertiesCfg(mesh_approximation="none"),
            articulation_props=ArticulationRootPropertiesCfg(
                articulation_enabled=False
            ),
            rigid_props=RigidBodyPropertiesCfg(rigid_body_enabled=False),
        ),  # type: ignore
        init_state=ArticulationCfg.InitialStateCfg(
            pos=(0.65, 0.0, 0.0),
            rot=rpy_to_quat((0.0, 0.0, 45.0)),
        ),
    )


@configclass
class EventCfg(ManipulationEventCfg):
    randomize_object_state: EventTermCfg = EventTermCfg(
        func=reset_root_state_uniform,
        mode="reset",
        params={
            "asset_cfg": SceneEntityCfg("bolt"),
            "pose_range": {},
            "velocity_range": {},
        },
    )


@configclass
class TaskCfg(ManipulationEnvCfg):
    ## Scenario
    domain: Domain = Domain.MARS

    ## Assets
    robot: Manipulator | AssetVariant = assets.Franka(
        end_effector=assets.ElectricScrewdriverM5()
    )
    bolt: Object | AssetVariant | None = assets.BoltM8()

    ## Scene
    scene: SceneCfg = SceneCfg()

    ## Events
    events: EventCfg = EventCfg()

    ## Time
    env_rate: float = 1.0 / 500.0
    episode_length_s: float = 15.0
    is_finite_horizon: bool = True

    def __post_init__(self):
        super().__post_init__()

        # Task setup: Nut from Bolt
        if isinstance(self.bolt, assets.BoltM8):
            nut = assets.NutM8()
            thread_offset = -0.000375
            thread_pitch = 0.00125
        else:
            raise ValueError("Unsupported bolt type")
        # Scene: Nut
        self.scene.nut = nut.asset_cfg
        self.scene.nut.prim_path = "{ENV_REGEX_NS}/nut"
        self.scene.nut.init_state.pos = (
            0.65,
            0.0,
            0.493,
        )
        # Scene: Bolt
        self.scene.bolt = self.bolt.asset_cfg
        self.scene.bolt.prim_path = "{ENV_REGEX_NS}/bolt"
        self.scene.bolt.init_state.pos = (
            0.65,
            0.0,
            0.493 + thread_offset + 3.0 * thread_pitch,
        )
        # Scene: Decor
        self.scene.decor.prim_path = (
            "/World/decor" if self.stack else "{ENV_REGEX_NS}/decor"
        )

        # Sensor: End-effector contacts
        if isinstance(self.scene.contacts_end_effector, ContactSensorCfg):
            self.scene.contacts_end_effector.filter_prim_paths_expr = [
                self.scene.bolt.prim_path
            ]


############
### Task ###
############


class Task(ManipulationEnv):
    cfg: TaskCfg

    def __init__(self, cfg: TaskCfg, **kwargs):
        super().__init__(cfg, **kwargs)

        ## Get scene assets
        self._obj: RigidObject = self.scene["bolt"]
        self._target: RigidObject = self.scene["nut"]

        ## Initialize buffers
        self._offset_pos_bolt_driver_slot = torch.tensor(
            self.cfg.bolt.frame_driver_slot.offset.pos,  # type: ignore
            dtype=torch.float32,
            device=self.device,
        ).repeat(self.num_envs, 1)

    def _reset_idx(self, env_ids: Sequence[int]):
        super()._reset_idx(env_ids)

    def extract_step_return(self) -> StepReturn:
        return _compute_step_return(
            ## Time
            episode_length=self.episode_length_buf,
            max_episode_length=self.max_episode_length,
            truncate_episodes=self.cfg.truncate_episodes,
            ## Actions
            act_current=self.action_manager.action,
            act_previous=self.action_manager.prev_action,
            ## States
            # Joints
            joint_pos_robot=self._robot.data.joint_pos,
            joint_pos_limits_robot=(
                self._robot.data.soft_joint_pos_limits
                if torch.all(torch.isfinite(self._robot.data.soft_joint_pos_limits))
                else None
            ),
            joint_pos_end_effector=self._end_effector.data.joint_pos
            if isinstance(self._end_effector, Articulation)
            else None,
            joint_pos_limits_end_effector=(
                self._end_effector.data.soft_joint_pos_limits
                if isinstance(self._end_effector, Articulation)
                and torch.all(
                    torch.isfinite(self._end_effector.data.soft_joint_pos_limits)
                )
                else None
            ),
            joint_acc_robot=self._robot.data.joint_acc,
            joint_applied_torque_robot=self._robot.data.applied_torque,
            # Kinematics
            fk_pos_end_effector=self._tf_end_effector.data.target_pos_source[:, 0, :],
            fk_quat_end_effector=self._tf_end_effector.data.target_quat_source[:, 0, :],
            # Transforms (world frame)
            tf_pos_end_effector=self._tf_end_effector.data.target_pos_w[:, 0, :],
            tf_quat_end_effector=self._tf_end_effector.data.target_quat_w[:, 0, :],
            tf_pos_obj=self._obj.data.root_pos_w,
            tf_quat_obj=self._obj.data.root_quat_w,
            tf_pos_target=self._target.data.root_pos_w,
            tf_quat_target=self._target.data.root_quat_w,
            # Object velocity
            vel_ang_obj=self._obj.data.root_ang_vel_w,
            # Contacts
            contact_forces_robot=self._contacts_robot.data.net_forces_w,  # type: ignore
            contact_forces_end_effector=self._contacts_end_effector.data.net_forces_w
            if isinstance(self._contacts_end_effector, ContactSensor)
            else None,
            contact_force_matrix_end_effector=self._contacts_end_effector.data.force_matrix_w
            if isinstance(self._contacts_end_effector, ContactSensor)
            else None,
            ## Bolt descriptors
            offset_pos_bolt_driver_slot=self._offset_pos_bolt_driver_slot,
        )


@torch.jit.script
def _compute_step_return(
    *,
    ## Time
    episode_length: torch.Tensor,
    max_episode_length: int,
    truncate_episodes: bool,
    ## Actions
    act_current: torch.Tensor,
    act_previous: torch.Tensor,
    ## States
    # Joints
    joint_pos_robot: torch.Tensor,
    joint_pos_limits_robot: torch.Tensor | None,
    joint_pos_end_effector: torch.Tensor | None,
    joint_pos_limits_end_effector: torch.Tensor | None,
    joint_acc_robot: torch.Tensor,
    joint_applied_torque_robot: torch.Tensor,
    # Kinematics
    fk_pos_end_effector: torch.Tensor,
    fk_quat_end_effector: torch.Tensor,
    # Transforms (world frame)
    tf_pos_end_effector: torch.Tensor,
    tf_quat_end_effector: torch.Tensor,
    tf_pos_obj: torch.Tensor,
    tf_quat_obj: torch.Tensor,
    tf_pos_target: torch.Tensor,
    tf_quat_target: torch.Tensor,
    # Object velocity
    vel_ang_obj: torch.Tensor,
    # Contacts
    contact_forces_robot: torch.Tensor,
    contact_forces_end_effector: torch.Tensor | None,
    contact_force_matrix_end_effector: torch.Tensor | None,
    ## Bolt descriptors
    offset_pos_bolt_driver_slot: torch.Tensor,
) -> StepReturn:
    num_envs = episode_length.size(0)
    dtype = episode_length.dtype
    device = episode_length.device

    ############
    ## States ##
    ############
    ## Joints
    # Robot joints
    joint_pos_robot_normalized = (
        scale_transform(
            joint_pos_robot,
            joint_pos_limits_robot[:, :, 0],
            joint_pos_limits_robot[:, :, 1],
        )
        if joint_pos_limits_robot is not None
        else joint_pos_robot
    )
    # End-effector joints
    joint_pos_end_effector_normalized = (
        scale_transform(
            joint_pos_end_effector,
            joint_pos_limits_end_effector[:, :, 0],
            joint_pos_limits_end_effector[:, :, 1],
        )
        if joint_pos_end_effector is not None
        and joint_pos_limits_end_effector is not None
        else (
            joint_pos_end_effector
            if joint_pos_end_effector is not None
            else torch.empty((num_envs, 0), dtype=dtype, device=device)
        )
    )

    ## Kinematics
    fk_rotmat_end_effector = matrix_from_quat(fk_quat_end_effector)
    fk_rot6d_end_effector = rotmat_to_rot6d(fk_rotmat_end_effector)

    ## Transforms (world frame)
    # End-effector -> Bolt driver slot
    tf_pos_bolt_driver_slot, tf_quat_bolt_driver_slot = combine_frame_transforms(
        t01=tf_pos_obj,
        q01=tf_quat_obj,
        t12=offset_pos_bolt_driver_slot,
    )
    (
        tf_pos_end_effector_to_bolt_driver_slot,
        tf_quat_end_effector_to_bolt_driver_slot,
    ) = subtract_frame_transforms(
        t01=tf_pos_end_effector,
        q01=tf_quat_end_effector,
        t02=tf_pos_bolt_driver_slot,
        q02=tf_quat_bolt_driver_slot,
    )
    tf_rotmat_end_effector_to_bolt_driver_slot = matrix_from_quat(
        tf_quat_end_effector_to_bolt_driver_slot
    )
    tf_rot6d_end_effector_to_bolt_driver_slot = rotmat_to_rot6d(
        tf_rotmat_end_effector_to_bolt_driver_slot
    )
    dist_end_effector_to_bolt_driver_slot = torch.norm(
        tf_pos_end_effector_to_bolt_driver_slot, dim=-1
    )
    # Object -> Target
    tf_pos_obj_to_target, tf_quat_obj_to_target = subtract_frame_transforms(
        t01=tf_pos_obj,
        q01=tf_quat_obj,
        t02=tf_pos_target,
        q02=tf_quat_target,
    )
    tf_rotmat_obj_to_target = matrix_from_quat(tf_quat_obj_to_target)
    tf_rot6d_obj_to_target = rotmat_to_rot6d(tf_rotmat_obj_to_target)

    ## Contacts
    contact_forces_mean_robot = contact_forces_robot.mean(dim=1)
    contact_forces_mean_end_effector = (
        contact_forces_end_effector.mean(dim=1)
        if contact_forces_end_effector is not None
        else torch.empty((num_envs, 0), dtype=dtype, device=device)
    )
    contact_forces_end_effector = (
        contact_forces_end_effector
        if contact_forces_end_effector is not None
        else torch.empty((num_envs, 0), dtype=dtype, device=device)
    )

    #############
    ## Rewards ##
    #############
    # Penalty: Action rate
    WEIGHT_ACTION_RATE = -0.05
    penalty_action_rate = WEIGHT_ACTION_RATE * torch.sum(
        torch.square(act_current - act_previous), dim=1
    )

    # Penalty: Joint torque
    WEIGHT_JOINT_TORQUE = -0.00025
    MAX_JOINT_TORQUE_PENALTY = -4.0
    penalty_joint_torque = torch.clamp_min(
        WEIGHT_JOINT_TORQUE
        * torch.sum(torch.square(joint_applied_torque_robot), dim=1),
        min=MAX_JOINT_TORQUE_PENALTY,
    )

    # Penalty: Joint acceleration
    WEIGHT_JOINT_ACCELERATION = -0.005
    MAX_JOINT_ACCELERATION_PENALTY = -4.0
    penalty_joint_acceleration = torch.clamp_min(
        WEIGHT_JOINT_ACCELERATION * torch.sum(torch.square(joint_acc_robot), dim=1),
        min=MAX_JOINT_ACCELERATION_PENALTY,
    )

    # Penalty: Undesired robot contacts
    WEIGHT_UNDESIRED_ROBOT_CONTACTS = -1.0
    THRESHOLD_UNDESIRED_ROBOT_CONTACTS = 10.0
    penalty_undesired_robot_contacts = WEIGHT_UNDESIRED_ROBOT_CONTACTS * (
        torch.max(torch.norm(contact_forces_robot, dim=-1), dim=1)[0]
        > THRESHOLD_UNDESIRED_ROBOT_CONTACTS
    )

    # Reward: End-effector top-down orientation
    WEIGHT_TOP_DOWN_ORIENTATION = 8.0
    TANH_STD_TOP_DOWN_ORIENTATION = 0.025
    top_down_alignment = torch.sum(
        fk_rotmat_end_effector[:, :, 2]
        * torch.tensor((0.0, 0.0, -1.0), device=device)
        .unsqueeze(0)
        .expand(num_envs, 3),
        dim=1,
    )
    reward_top_down_orientation = WEIGHT_TOP_DOWN_ORIENTATION * (
        1.0 - torch.tanh((1.0 - top_down_alignment) / TANH_STD_TOP_DOWN_ORIENTATION)
    )

    # Reward: Object upright orientation
    WEIGHT_OBJECT_TOP_DOWN_ORIENTATION = -64.0
    TANH_STD_OBJECT_TOP_DOWN_ORIENTATION = 0.05
    object_top_down_alignment = torch.sum(
        matrix_from_quat(tf_quat_obj)[:, :, 2]
        * torch.tensor((0.0, 0.0, 1.0), device=device).unsqueeze(0).expand(num_envs, 3),
        dim=1,
    )
    penalty_object_top_down_orientation = WEIGHT_OBJECT_TOP_DOWN_ORIENTATION * (
        torch.tanh(
            (1.0 - object_top_down_alignment) / TANH_STD_OBJECT_TOP_DOWN_ORIENTATION
        )
    )

    # Reward: Distance | End-effector <--> Object
    WEIGHT_DISTANCE_END_EFFECTOR_TO_BOLT_DRIVER_SLOT = 12.0
    TANH_STD_DISTANCE_END_EFFECTOR_TO_BOLT_DRIVER_SLOT = 0.4
    reward_distance_end_effector_to_bolt_driver_slot = (
        WEIGHT_DISTANCE_END_EFFECTOR_TO_BOLT_DRIVER_SLOT
        * (
            1.0
            - torch.tanh(
                dist_end_effector_to_bolt_driver_slot
                / TANH_STD_DISTANCE_END_EFFECTOR_TO_BOLT_DRIVER_SLOT
            )
        )
    )

    # Reward: Distance | End-effector <--> Object
    WEIGHT_DISTANCE_END_EFFECTOR_TO_BOLT_DRIVER_SLOT_CLOSE = 32.0
    TANH_STD_DISTANCE_END_EFFECTOR_TO_BOLT_DRIVER_SLOT_CLOSE = 0.02
    reward_distance_end_effector_to_bolt_driver_slot_close = (
        WEIGHT_DISTANCE_END_EFFECTOR_TO_BOLT_DRIVER_SLOT_CLOSE
        * (
            1.0
            - torch.tanh(
                dist_end_effector_to_bolt_driver_slot
                / TANH_STD_DISTANCE_END_EFFECTOR_TO_BOLT_DRIVER_SLOT_CLOSE
            )
        )
    )

    # Reward: Distance | End-effector <--> Object
    WEIGHT_DISTANCE_END_EFFECTOR_TO_BOLT_DRIVER_SLOT_PRECISION = 512.0
    TANH_STD_DISTANCE_END_EFFECTOR_TO_BOLT_DRIVER_SLOT_PRECISION = 0.005
    reward_distance_end_effector_to_bolt_driver_slot_precision = (
        WEIGHT_DISTANCE_END_EFFECTOR_TO_BOLT_DRIVER_SLOT_PRECISION
        * (
            1.0
            - torch.tanh(
                dist_end_effector_to_bolt_driver_slot
                / TANH_STD_DISTANCE_END_EFFECTOR_TO_BOLT_DRIVER_SLOT_PRECISION
            )
        )
    )

    # Reward: Contact object
    WEIGHT_CONTACT = 4.0
    THRESHOLD_CONTACT = 5.0
    reward_contact = (
        WEIGHT_CONTACT
        * (
            torch.mean(
                torch.max(
                    torch.norm(contact_force_matrix_end_effector, dim=-1), dim=-1
                )[0],
                dim=1,
            )
            > THRESHOLD_CONTACT
        )
        if contact_force_matrix_end_effector is not None
        else torch.zeros(num_envs, dtype=dtype, device=device)
    )

    # Reward: Screwing based on angular velocity of the object
    WEIGHT_SCREWING = 128.0
    reward_screwing = WEIGHT_SCREWING * (-vel_ang_obj[:, 2] / torch.pi)

    # Penalty: Distance | Object <--> Target
    WEIGHT_DISTANCE_OBJ_TO_TARGET_TOO_FAR = -256.0
    THRESHOLD_DISTANCE_OBJ_TO_TARGET_TOO_FAR = 0.075
    is_obj_too_far = (
        torch.norm(tf_pos_obj_to_target[:, :2], dim=-1)
        > THRESHOLD_DISTANCE_OBJ_TO_TARGET_TOO_FAR
    )
    penalty_distance_obj_to_target_too_far = (
        WEIGHT_DISTANCE_OBJ_TO_TARGET_TOO_FAR * is_obj_too_far.float()
    )

    # Reward: Distance | Object <--> Target
    WEIGHT_DISTANCE_OBJ_TO_TARGET = 2048.0
    TANH_STD_DISTANCE_OBJ_TO_TARGET = 0.00075
    reward_distance_obj_to_target = (
        WEIGHT_DISTANCE_OBJ_TO_TARGET
        * (~is_obj_too_far).float()
        * (
            1.0
            - torch.tanh(
                torch.clip(-tf_pos_obj_to_target[:, 2], 0.0)
                / TANH_STD_DISTANCE_OBJ_TO_TARGET
            )
        )
    )

    ##################
    ## Terminations ##
    ##################
    # No termination condition
    termination = torch.zeros(num_envs, dtype=torch.bool, device=device)
    # Truncation
    truncation = (
        episode_length >= max_episode_length
        if truncate_episodes
        else torch.zeros(num_envs, dtype=torch.bool, device=device)
    )

    return StepReturn(
        {
            "state": {
                "vel_ang_obj": vel_ang_obj,
                "contact_forces_mean_robot": contact_forces_mean_robot,
                "contact_forces_mean_end_effector": contact_forces_mean_end_effector,
                "tf_pos_end_effector_to_bolt_driver_slot": tf_pos_end_effector_to_bolt_driver_slot,
                "tf_rot6d_end_effector_to_bolt_driver_slot": tf_rot6d_end_effector_to_bolt_driver_slot,
                "tf_pos_obj_to_target": tf_pos_obj_to_target,
                "tf_rot6d_obj_to_target": tf_rot6d_obj_to_target,
            },
            "state_dyn": {
                "contact_forces_robot": contact_forces_robot,
                "contact_forces_end_effector": contact_forces_end_effector,
            },
            "proprio": {
                "fk_pos_end_effector": fk_pos_end_effector,
                "fk_rot6d_end_effector": fk_rot6d_end_effector,
            },
            "proprio_dyn": {
                "joint_pos_robot_normalized": joint_pos_robot_normalized,
                "joint_pos_end_effector_normalized": joint_pos_end_effector_normalized,
                "joint_acc_robot": joint_acc_robot,
                "joint_applied_torque_robot": joint_applied_torque_robot,
            },
        },
        {
            "penalty_action_rate": penalty_action_rate,
            "penalty_joint_torque": penalty_joint_torque,
            "penalty_joint_acceleration": penalty_joint_acceleration,
            "penalty_undesired_robot_contacts": penalty_undesired_robot_contacts,
            "reward_top_down_orientation": reward_top_down_orientation,
            "penalty_object_top_down_orientation": penalty_object_top_down_orientation,
            "reward_distance_end_effector_to_bolt_driver_slot": reward_distance_end_effector_to_bolt_driver_slot,
            "reward_distance_end_effector_to_bolt_driver_slot_close": reward_distance_end_effector_to_bolt_driver_slot_close,
            "reward_distance_end_effector_to_bolt_driver_slot_precision": reward_distance_end_effector_to_bolt_driver_slot_precision,
            "reward_contact": reward_contact,
            "reward_screwing": reward_screwing,
            "penalty_distance_obj_to_target_too_far": penalty_distance_obj_to_target_too_far,
            "reward_distance_obj_to_target": reward_distance_obj_to_target,
        },
        termination,
        truncation,
    )



from dataclasses import MISSING

import torch

from srb import assets
from srb.core.action import WheeledDriveAction
from srb.core.asset import AssetVariant, GroundRobot
from srb.core.env import ViewerCfg
from srb.core.env.mobile.env import (
    MobileEnv,
    MobileEnvCfg,
    MobileEventCfg,
    MobileSceneCfg,
)
from srb.core.manager import EventTermCfg, SceneEntityCfg
from srb.core.mdp import reset_root_state_uniform
from srb.utils.cfg import configclass
from srb.utils.math import deg_to_rad


@configclass
class GroundSceneCfg(MobileSceneCfg):
    env_spacing: float = 32.0


@configclass
class GroundEventCfg(MobileEventCfg):
    randomize_robot_state: EventTermCfg = EventTermCfg(
        func=reset_root_state_uniform,
        mode="reset",
        params={
            "asset_cfg": SceneEntityCfg("robot"),
            "pose_range": {
                "x": (-0.5, 0.5),
                "y": (-0.5, 0.5),
                "z": (0.4, 0.6),
                "yaw": (-torch.pi, torch.pi),
            },
            "velocity_range": {
                "x": (-0.5, 0.5),
                "y": (-0.5, 0.5),
                "z": (0.0, 0.5),
                "roll": (-deg_to_rad(5.0), deg_to_rad(5.0)),
                "pitch": (-deg_to_rad(5.0), deg_to_rad(5.0)),
                "yaw": (-deg_to_rad(15.0), deg_to_rad(15.0)),
            },
        },
    )


@configclass
class GroundEnvCfg(MobileEnvCfg):
    ## Assets
    robot: GroundRobot | AssetVariant = assets.Perseverance()
    _robot: GroundRobot = MISSING  # type: ignore

    ## Scene
    scene: GroundSceneCfg = GroundSceneCfg()

    ## Events
    events: GroundEventCfg = GroundEventCfg()

    ## Time
    env_rate: float = 1.0 / 50.0
    agent_rate: float = 1.0 / 25.0

    ## Viewer
    viewer: ViewerCfg = ViewerCfg(
        eye=(7.5, -7.5, 15.0), lookat=(0.0, 0.0, 0.0), origin_type="env"
    )

    def __post_init__(self):
        super().__post_init__()


class GroundEnv(MobileEnv):
    cfg: GroundEnvCfg

    def __init__(self, cfg: GroundEnvCfg, **kwargs):
        super().__init__(cfg, **kwargs)

        ## Check if the robot uses a wheeled drive action term
        self._wheeled_drive_action_term_key = next(
            filter(
                lambda term_key: isinstance(
                    self.action_manager._terms[term_key], WheeledDriveAction
                ),
                self.action_manager._terms.keys(),
            ),
            None,
        )
        self._forward_drive_indices = (
            [0] if self._wheeled_drive_action_term_key is not None else []
        )



from srb.utils.registry import register_srb_tasks

from .task import Task, TaskCfg
from .task_multi import MultiTask, MultiTaskCfg
from .task_visual import VisualMultiTask, VisualMultiTaskCfg, VisualTask, VisualTaskCfg

BASE_TASK_NAME = __name__.split(".")[-1]
register_srb_tasks(
    {
        BASE_TASK_NAME: {},
        f"{BASE_TASK_NAME}_visual": {
            "entry_point": VisualTask,
            "task_cfg": VisualTaskCfg,
        },
        f"multi_{BASE_TASK_NAME}": {
            "entry_point": MultiTask,
            "task_cfg": MultiTaskCfg,
        },
        f"multi_{BASE_TASK_NAME}_visual": {
            "entry_point": VisualMultiTask,
            "task_cfg": VisualMultiTaskCfg,
        },
    },
    default_entry_point=Task,
    default_task_cfg=TaskCfg,
)



from .anymal import AnymalC, AnymalD, RandomAnymalQuadruped  # noqa: F401
from .cassie import Cassie  # noqa: F401
from .crazyflie import Crazyflie  # noqa: F401
from .ingenuity import Ingenuity  # noqa: F401
from .lander import (  # noqa: F401
    ApolloLander,
    PeregrineLander,
    RandomLander,
    ResilienceLander,
    VikramLander,
)
from .leo_rover import LeoRover  # noqa: F401
from .mars_rover import MarsRover  # noqa: F401
from .perseverance import Perseverance  # noqa: F401
from .pragyan import Pragyan  # noqa: F401
from .spacecraft import (  # noqa: F401
    Cubesat,
    Gateway,
    Starship,
    SuperHeavy,
    VenusExpress,
)
from .spot import Spot  # noqa: F401
from .unitree import (  # noqa: F401
    RandomUnitreeQuadruped,
    UnitreeA1,
    UnitreeGo1,
    UnitreeGo2,
)



from srb.core.asset import AssetBaseCfg, Subterrane, Terrain
from srb.core.sim import CollisionPropertiesCfg, GridParticlesSpawnerCfg, UsdFileCfg
from srb.utils.path import SRB_ASSETS_DIR_SRB_SCENERY


class LunaLab(Subterrane):
    asset_cfg: AssetBaseCfg = AssetBaseCfg(
        prim_path="{ENV_REGEX_NS}/lunalab",
        spawn=UsdFileCfg(
            usd_path=SRB_ASSETS_DIR_SRB_SCENERY.joinpath("lunalab.usdc").as_posix(),
            collision_props=CollisionPropertiesCfg(),
        ),
        init_state=AssetBaseCfg.InitialStateCfg(pos=(-4.25, -5.5, 0.0)),
    )

    PARTICLE_SIZE: float = 0.025
    _regolith: AssetBaseCfg = AssetBaseCfg(
        prim_path="{ENV_REGEX_NS}/regolith",
        spawn=GridParticlesSpawnerCfg(
            ratio=0.5,
            particle_size=PARTICLE_SIZE,
            dim_x=round(6.5 / PARTICLE_SIZE),
            dim_y=round(11.0 / PARTICLE_SIZE),
            dim_z=round(0.5 / PARTICLE_SIZE),
            velocity=((-0.5, 0.5), (-0.5, 0.5), (-0.5, 0.0)),
            fluid=False,
            density=1500.0,
            friction=0.85,
            cohesion=0.65,
        ),
        init_state=AssetBaseCfg.InitialStateCfg(pos=(0.0, 0.0, -0.3)),
    )


class Oberpfaffenhofen(Terrain):
    asset_cfg: AssetBaseCfg = AssetBaseCfg(
        prim_path="{ENV_REGEX_NS}/oberpfaffenhofen",
        spawn=UsdFileCfg(
            usd_path=SRB_ASSETS_DIR_SRB_SCENERY.joinpath(
                "oberpfaffenhofen_test_site.usdc"
            ).as_posix(),
            collision_props=CollisionPropertiesCfg(),
        ),
    )
